--------------------------------------------------------------------------------
-- SubModule DisplayDriver
-- Created   3/3/2009 11:14:26 AM
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity DisplayDriver is port
   (
     RST        : in    std_logic;
     CLK        : in    std_logic;
     ONEPPS     : in    std_logic;
     CNT        : in    std_logic_vector(25 downto 0);
     MESS       : in    std_logic_vector(79 downto 0);
     ENABLE     : in    std_logic;
     IS422      : in    std_logic;
     BCD        : out   std_logic_vector (39 downto 0);

     BUSY       : in    std_logic;
     LD         : out   std_logic;
     DOUT       : out   std_logic_vector(7 downto 0);

     RDY        : in    std_logic;
     ERR        : in    std_logic;
     DIN        : in    std_logic_vector(7 downto 0);

     SEND       : out   std_logic
   );
end DisplayDriver;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of DisplayDriver is

-- Type for constants
   type digit is range 0 to 255;
   type digit_array is array (natural range <>) of digit;

-- Constant Declarations

   -- Configuring the display
   constant ini : digit_array := 1 &    -- start message                   0x01     [fixed, pre-defined value of 1]
                                 0 &    -- destination address             0x00     [set to 0, only broadcast messages are displayed]
                                 255 &  -- source address                  0xFF     [fixed, pre-defined value of 255]
                                 1 &    -- protocol version                0x01     [fixed value]
                                 8 &    -- message length                  0x08     [counts from the next ("message ID") byte until (but not including!) the CRC]
                                 1 &    -- message ID                      0x01     [01 for "configure the display" type of message]
                                 90 &   -- display width in pixels         0x3C     [sets effective display width]
                                 7 &    -- display height in pixels        0x07     [sets effective display height]
                                 2 &    -- display type                    0x01     [2 for displaying TRANSIENT messages]
                                 1 &    -- number of screens               0x01     [sets the number of screens used; 17 is the maximum]
                                 1 &    -- number of frames per screen     0x01
                                 -- 255 &  -- duration for screen             0xFF     [sets the duration for the screen, in miliseconds]
                                 -- 255;   -- duration for screen             0xFF
                                 3 &       -- 1000ms duration = 0x03E8
                                 232;      -- 0x03 = 3; 0xE8 = 232
            -- This message looks like:
            -- 01 00 FF 01 08 01 5A 07 02 01 01 03 E8 <CRC>
            -- The response should be like
            -- 01 FF 01 01 03 01 00 06 <CRC>

   constant ini_len : positive := ini'length(1);

   -- Start the display
   constant shw : digit_array := 1 &    -- start message                   0x01     [fixed, pre-defined value of 1]
                                 0 &    -- destination address             0x00     [set to 0, only broadcast messages are displayed]
                                 255 &  -- source address                  0xFF     [fixed, pre-defined value of 255]
                                 1 &    -- protocol version                0x01     [fixed value]
                                 2 &    -- message length                  0x02     [counts from the next ("message ID") byte until (but not including!) the CRC]
                                 4 &    -- message ID                      0x04     [o4 for "start the display" command]
                                 2;     -- display type                    0x01     [2 for displaying TRANSIENT transient messages]
            -- This message looks like:
            -- 01 00 FF 01 02 04 02 <CRC>
            -- The response should be like
            -- 01 FF 01 01 03 04 00 06 <CRC>

   constant shw_len : positive := shw'length(1);

   -- Loads the string into the display memory
   constant msg : digit_array := 1 &    -- start message                   0x01     [fixed, pre-defined value of 1]
                                 0 &    -- destination address             0x00     [set to 0, only broadcast messages are displayed]
                                 255 &  -- source address                  0xFF     [fixed, pre-defined value of 255]
                                 1 &    -- protocol version                0x01     [fixed value]
                                 24 &   -- message length                  0x18     [counts from the next ("message ID") byte until (but not including!) the CRC]
                                 2 &    -- message ID                      0x02     [02 for "load the string" type of message]
                                 1 &    -- font number                     0x00     [for now, set to 1]                      **********
                                 2 &    -- display type                    0x01     [2 for displaying TRANSIENT messages]
                                 1 &    -- screen number                   0x01     [here, screen #1 is used; 17 screens total]
                                 1 &    -- number of frames                0x01     []
                                 7 &    -- number of rows per frame        0x07     []
                                 10 &   -- string length                   0x0A     [here, 10 characters for displaying GPS-second information]
                                 32&32&32&32&32&
                                 32&32&32&32&32& -- string (10 char)      [0x20]
                                 191 &  -- row 1: bright green             0xBF     [defines the color of the row]
                                 191 &  -- row 2                           0xBF
                                 191 &  -- row 3                           0xBF
                                 191 &  -- row 4                           0xBF
                                 191 &  -- row 5                           0xBF
                                 191 &  -- row 6                           0xBF
                                 191;   -- row 7                           0xBF
            -- This message looks like:
            -- 01 00 FF 01 18 02 01 02 01 01 07 0A 20 20 20 20
            -- 20 20 20 20 20 20 BF BF BF BF BF BF BF <CRC>
            -- The response should be like
            -- 01 FF 01 01 03 02 00 06 <CRC>

   constant msg_len : positive := msg'length(1);

-- Component Declarations
--   component BCDConv
--      generic (
--         M           : positive;      -- number of binary digits
--         N           : positive       -- number of BCD digits
--      );
--      port (
--         CLK         : in  std_logic;
--         RST         : in  std_logic;
--         INIT        : in  std_logic;  -- initialize conversion
--         BUSY        : out std_logic;  -- indicates busy with conversion
--         DIN         : in  std_logic_vector (M - 1 downto 0);    -- binary input
--         DOUT        : out std_logic_vector (4 * N -1 downto 0); -- BCD result
--         OVF         : out std_logic   -- overflow
--      );
--   end component;

-- Type Declarations
   type state_type is (IDLE, FAIL, INIT1, INIT2, MESSAGE, RESPONSE1,
                       RESPONSE2, RESPONSE3, RESPONSE4, WAITING);

-- Signal Declarations
   signal state              : state_type;
   signal next_state         : state_type;
   signal onepps_old         : std_logic;
   signal onepps_trig        : std_logic;
   signal count              : unsigned (7 downto 0);
   signal checksum           : std_logic_vector (7 downto 0);
   signal gps_bcd            : std_Logic_vector (39 downto 0);

-- Functions
   function get_byte (s : in digit_array; i : positive) return std_logic_vector is
      variable c : natural;
   begin
      c := natural (s (i-1));
      return std_logic_vector (To_unsigned (c, 8));
   end function get_byte;

begin

   -- trigger on positive edge of 1 PPS
   onepps_old <= onepps when rising_edge (CLK);
   onepps_trig <= onepps and not onepps_old;

   -- state machine
   fsm: process (CLK) is
      variable byte : std_logic_vector (7 downto 0);
   begin
      if rising_edge (CLK) then
         LD <= '0';
         if (RST = '1') or (ENABLE = '0') then
            state <= FAIL;
            SEND <= '0';
         else
            case state is
               -- fail state: need to initialize display
               when FAIL =>
                  SEND <= '0';
                  if onepps_trig = '1' then
                     count <= To_Unsigned (1, 8);
                     checksum <= (others => '0');
                     state <= INIT1;
                  end if;

               -- wait for next 1 PPS to send GPS number
               when IDLE =>
                  SEND <= '0';
                  if onepps_trig = '1' then
                     count <= To_Unsigned (1, 8);
                     checksum <= (others => '0');
                     state <= MESSAGE;
                  end if;

               -- send display configuration
               when INIT1 =>
                  SEND <= '1';
                  if BUSY = '0' then
                     if count /= To_unsigned (ini_len+1, 8) then
                        byte := get_byte (ini, To_integer (count));
                     else
                        byte := checksum;
                     end if;
                     DOUT <= byte;
                     checksum <= checksum xor byte;
                     LD <= '1';
                     count <= count + 1;
                     if count = To_unsigned (ini_len+1, 8) then
                        checksum <= (others => '0');
                        state <= RESPONSE1;
                        next_state <= INIT2;
                     else
                        state <= WAITING;
                        next_state <= INIT1;
                     end if;
                  end if;

               -- show screen
               when INIT2 =>
                  SEND <= '1';
                  if BUSY = '0' then
                     if count /= To_unsigned (shw_len+1, 8) then
                        byte := get_byte (shw, To_integer (count));
                     else
                        byte := checksum;
                     end if;
                     DOUT <= byte;
                     checksum <= checksum xor byte;
                     LD <= '1';
                     count <= count + 1;
                     if count = To_unsigned (shw_len+1, 8) then
                        checksum <= (others => '0');
                        state <= RESPONSE1;
                        next_state <= IDLE;
                     else
                        state <= WAITING;
                        next_state <= INIT2;
                     end if;
                  end if;

               -- send GPS message
               when MESSAGE =>
                  SEND <= '1';
                  if BUSY = '0' then
                     if    count = To_Unsigned (13, 8) then
                        byte := MESS(79 downto 72);
                     elsif count = To_Unsigned (14, 8) then
                        byte := MESS(71 downto 64);
                     elsif count = To_Unsigned (15, 8) then
                        byte := MESS(63 downto 56);
                     elsif count = To_Unsigned (16, 8) then
                        byte := MESS(55 downto 48);
                     elsif count = To_Unsigned (17, 8) then
                        byte := MESS(47 downto 40);
                     elsif count = To_Unsigned (18, 8) then
                        byte := MESS(39 downto 32);
                     elsif count = To_Unsigned (19, 8) then
                        byte := MESS(31 downto 24);
                     elsif count = To_Unsigned (20, 8) then
                        byte := MESS(23 downto 16);
                     elsif count = To_Unsigned (21, 8) then
                        byte := MESS(15 downto  8);
                     elsif count = To_Unsigned (22, 8) then
                        byte := MESS(7 downto   0);
                     elsif count /= To_unsigned (msg_len+1, 8) then
                        byte := get_byte (msg, To_integer (count));
                     else
                        byte := checksum;
                     end if;
                     DOUT <= byte;
                     checksum <= checksum xor byte;
                     LD <= '1';
                     count <= count + 1;
                     if count = To_unsigned (msg_len+1, 8) then
                        checksum <= (others => '0');
                        state <= RESPONSE1;
                        next_state <= IDLE;
                     else
                        state <= WAITING;
                        next_state <= MESSAGE;
                     end if;
                  end if;

               -- wait for response (only look for ack or nak & checksum)
               -- delay by one
               when RESPONSE1 =>
                  SEND <= '1';
                  state <= RESPONSE2;
               -- make sure last character has been sent
               when RESPONSE2 =>
                  SEND <= '1';
                  if BUSY = '0' then
                     state <= RESPONSE3;
                  end if;
                  count <= To_Unsigned (1, 8);
               -- start listening and wait for ACK or NAK
               when RESPONSE3 =>
                  SEND <= '0';
                  if (onepps = '1') and (onepps_old = '0') then
                     state <= FAIL;
                  elsif RDY = '1' then
                     if (DIN = X"06") or (DIN = X"15") then -- ack or nak
                        count <= To_Unsigned (255, 8);
                        state <= RESPONSE4;
                     elsif count = To_Unsigned (255, 8) then -- why so many?
                        state <= FAIL;
                     else
                        count <= count + 1;
                     end if;
                  end if;
               -- wait for one more (checksum)
               when RESPONSE4 =>
                  SEND <= '0';
                  if (onepps = '1') and (onepps_old = '0') then
                     state <= FAIL;
                  elsif RDY = '1' then
                     state <= next_state;
                  end if;
                  count <= To_Unsigned (1, 8);

               -- wait one cycle
               when WAITING =>
                  SEND <= '1';
                  state <= next_state;

            end case;
         end if;
      end if;
   end process fsm;

--   -- binary to BCD conversion
--   binBCDconv: BCDConv
--      generic map (
--         M => 32,
--         N => 10
--      )
--      port map (
--         CLK => CLK,
--         RST => RST,
--         INIT => onepps_trig,
--         BUSY => open,
--         DIN => GPS,
--         DOUT => gps_bcd,
--         OVF => open
--      );
--   BCD <= gps_bcd;


end Structure;
--------------------------------------------------------------------------------
