--------------------------------------------------------------------------------
-- SubModule DisplayDriverED
-- Created   3/3/2009 11:14:26 AM
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity DisplayDriverED is port
   (
     RST        : in    std_logic;
     CLK        : in    std_logic;
     ONEPPS     : in    std_logic;
     CNT        : in    std_logic_vector(25 downto 0);
     MESS       : in    std_logic_vector(79 downto 0);
     ENABLE     : in    std_logic;
     IS422      : in    std_logic;
     BCD        : out   std_logic_vector (39 downto 0);

     BUSY       : in    std_logic;
     LD         : out   std_logic;
     DOUT       : out   std_logic_vector(7 downto 0);

     RDY        : in    std_logic;
     ERR        : in    std_logic;
     DIN        : in    std_logic_vector(7 downto 0);

     SEND       : out   std_logic
   );
end DisplayDriverED;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of DisplayDriverED is

-- Type for constants
   type digit is range 0 to 255;
   type digit_array is array (natural range <>) of digit;

-- Constant Declarations

   -- Loads the string into the display memory
   constant msg : digit_array := 2 &    -- start message                   0x02     [STX/ctrl-B]
                                 48 &   -- destination address             0x30     [ASCII 0]
                                 48 &   -- destination address             0x30     [ASCII 0]
                                 32&32&32&32&32&
                                 32&32&32&32&32& -- string (10 char)       0x20     [ASCII space]
                                 3;     -- stop message                    0x03     [ETX/ctrl-C]
            -- This message looks like:
            -- 02 30 30 20 20 20 20 20 20 20 20 03

   constant msg_len : positive := msg'length(1);

-- Type Declarations
   type state_type is (IDLE, FAIL,  MESSAGE, RESPONSE1, RESPONSE2, WAITING);

-- Signal Declarations
   signal state              : state_type;
   signal next_state         : state_type;
   signal onepps_old         : std_logic;
   signal onepps_trig        : std_logic;
   signal count              : unsigned (7 downto 0);

-- Functions
   function get_byte (s : in digit_array; i : positive) return std_logic_vector is
      variable c : natural;
   begin
      c := natural (s (i-1));
      return std_logic_vector (To_unsigned (c, 8));
   end function get_byte;

begin

   -- trigger on positive edge of 1 PPS
   onepps_old <= onepps when rising_edge (CLK);
   onepps_trig <= onepps and not onepps_old;

   -- state machine
   fsm: process (CLK) is
      variable byte : std_logic_vector (7 downto 0);
   begin
      if rising_edge (CLK) then
         LD <= '0';
         if (RST = '1') or (ENABLE = '0') then
            state <= FAIL;
            SEND <= '0';
         else
            case state is
               -- fail state: need to initialize display
               when FAIL =>
                  SEND <= '0';
                  if onepps_trig = '1' then
                     count <= To_Unsigned (1, 8);
                     state <= IDLE;
                  end if;

               -- wait for next 1 PPS to send GPS number
               when IDLE =>
                  SEND <= '0';
                  if onepps_trig = '1' then
                     count <= To_Unsigned (1, 8);
                     state <= MESSAGE;
                  end if;

               -- send GPS message
               when MESSAGE =>
                  SEND <= '1';
                  if BUSY = '0' then
                     if    count = To_Unsigned ( 4, 8) then
                        byte := MESS(79 downto 72);
                     elsif count = To_Unsigned ( 5, 8) then
                        byte := MESS(71 downto 64);
                     elsif count = To_Unsigned ( 6, 8) then
                        byte := MESS(63 downto 56);
                     elsif count = To_Unsigned ( 7, 8) then
                        byte := MESS(55 downto 48);
                     elsif count = To_Unsigned ( 8, 8) then
                        byte := MESS(47 downto 40);
                     elsif count = To_Unsigned ( 9, 8) then
                        byte := MESS(39 downto 32);
                     elsif count = To_Unsigned (10, 8) then
                        byte := MESS(31 downto 24);
                     elsif count = To_Unsigned (11, 8) then
                        byte := MESS(23 downto 16);
                     elsif count = To_Unsigned (12, 8) then
                        byte := MESS(15 downto  8);
                     elsif count = To_Unsigned (13, 8) then
                        byte := MESS(7 downto   0);
                     else
                        byte := get_byte (msg, To_integer (count));
                     end if;
                     DOUT <= byte;
                     LD <= '1';
                     count <= count + 1;
                     if count = To_unsigned (msg_len, 8) then
                        state <= WAITING;
                        next_state <= RESPONSE1;
                     else
                        state <= WAITING;
                        next_state <= MESSAGE;
                     end if;
                  end if;

               -- wait for response
               -- make sure last character has been sent
               when RESPONSE1 =>
                  SEND <= '1';
                  if BUSY = '0' then
                     state <= RESPONSE2;
                  end if;
                  count <= To_Unsigned (1, 8);
               -- here we could read the response
               when RESPONSE2 =>
                  SEND <= '0';
                  state <= IDLE; -- don't wait for anything

               -- wait one cycle
               when WAITING =>
                  SEND <= '1';
                  state <= next_state;

            end case;
         end if;
      end if;
   end process fsm;

end Structure;
--------------------------------------------------------------------------------
