--------------------------------------------------------------------------------
-- Generate a fake clock that starts at an arbitrary time. Useful for testing
-- the pending leap second feature as well as the code which calculates the
-- time difference between the reference time and an input IRIG-B signal. The
-- arbitrary time is specified by a start time, in seconds, and a subsecond
-- difference relative to the reference PPS pulse, in clock cycles. A positive
-- subsecond difference corresponds to a fake PPS pulse which leads the
-- reference time by N clock cycles.
--
-- Ilya Belopolski
-- January 2012
--
-- Debugged, negative time delay compatibility added
--
-- Stefan Countryman
-- August 2011
--------------------------------------------------------------------------------

library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity FakeClock is port
   (
     TIME_START              : in    std_logic_vector (31 downto 0);
     DIFF                    : in    std_logic_vector (31 downto 0);
     PPS                     : in    std_logic;
     DELAY_PPS               : out   std_logic;
     RST                     : in    std_logic;
     CLK                     : in    std_logic;
     TIME_OUT                : out   std_logic_vector (31 downto 0);
     SIGN                    : out   std_logic;
     VIEW_DELAY              : out   std_logic_vector (31 downto 0)
   );
end FakeClock;

architecture ark of FakeClock is

   signal pps_old          : std_logic;
   signal pps_ancient      : std_logic;
   signal pps_rise         : std_logic;                      -- reference PPS signal
   signal timetime         : unsigned (31 downto 0);         -- output time

   signal signsign         : std_logic;

   signal diffRaw          : std_logic_vector (31 downto 0);
   signal delayUse         : signed (31 downto 0);

   signal delayCLK         : signed (31 downto 0);           -- countdown timer until fake PPS signal
   signal delay_pps_rise   : std_logic;
   signal countdown        : std_logic;

begin

     -- Synchronize PPS with CLK
     pps_old <= PPS when rising_edge (CLK);
     pps_ancient <= pps_old when rising_edge (CLK);
     pps_rise <= '1' when (pps_old = '1') and (pps_ancient = '0') else '0';

     -- Extract sign and delay information
     signsign <= DIFF(31);
     diffRaw <= DIFF;

     -- Implement the subsecond time difference as a delay of some number of
     -- clock cycles. In other words, after the reference PPS arrives, wait some
     -- number of clock cycles and then generate the fake PPS. If the subsecond
     -- difference is positive, then the fake PPS leads the reference PPS by N
     -- clock cycles, so delay the fake PPS by (1sec - N). If the subsecond
     -- difference is negative, delay the fake by N clock cycles.
     findDelay : process (CLK) is
     begin
          if rising_edge (CLK) then
             if diffRaw = X"00000000" then
                delayUse <= X"00000000";
             elsif signsign = '1' then
                delayUse <= X"00000000" - signed(diffRaw);
             else
                delayUse <= 67108864 - signed(diffRaw);
             end if;
          end if;
     end process findDelay;

     -- Generate the delayed PPS signal
     delayClock : process (CLK) is
     begin
          if rising_edge (CLK) then

             if delayUse = 0 then
                delay_pps_rise <= pps_rise;
             elsif delay_pps_rise = '1' then
                delay_pps_rise <= '0';
             end if;

             if delayCLK = 0 and delayUse /= 0 then
                delay_pps_rise <= '1';
                countdown <= '0';
                delayCLK <= delayUse;
             elsif (pps_rise = '1' or countdown = '1') and delayUse /=0 then
                delayCLK <= delayCLK - 1;
                countdown <= '1';
             end if;

         end if;
     end process delayClock;

     -- Clock away
     clocky : process (CLK) is
     begin
          if rising_edge (CLK) then
             if RST = '1' then
                  timetime <= unsigned(TIME_START);
             elsif delay_pps_rise = '1' then
                timetime <= timetime + 1;
             end if;
          end if;
     end process clocky;

     TIME_OUT <= std_logic_vector(timetime);
     DELAY_PPS <= std_logic(delay_pps_rise);
     VIEW_DELAY <= std_logic_vector(DelayUse);
     SIGN <= signsign;

end ark;

