--------------------------------------------------------------------------------
-- SubModule IRIGB_clock
-- Created   2/2/2009 1:30:50 PM
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity IRIGB_clock is port
   (
     CLK       : in    std_logic;
     ONEPPS    : in    std_logic;
     POSMARK   : out   std_logic;
     SYMMARK   : out   std_logic;
     POSCNT    : out   std_logic_vector(3 downto 0);
     SYMCNT    : out   std_logic_vector(3 downto 0);
     POSEND    : out   std_logic;
     ZEROEND   : out   std_logic;
     ONEEND    : out   std_logic
   );
end IRIGB_clock;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of IRIGB_clock is

-- Constants
   constant divisor     : integer := 67108; -- 2**26 / 1000;
   constant numerator   : integer := 108; -- (2**26 % 1000)* denumerator/1000
   constant denumerator : integer := 125;

-- Component Declarations

-- Signal Declarations
   signal pps_old       : std_logic;  -- old 1 PPS
   signal pps_short     : std_logic;  -- a short version of the 1 PPS
   signal cnt           : unsigned (23 downto 0);  -- down counter
   signal remain        : signed (15 downto 0); -- remainder
   signal addcycle      : std_logic; -- add a clock cycle
   signal kHz           : std_logic; -- 1 kHz clock enable
   signal pos_cnt       : unsigned (3 downto 0); -- position marker index
   signal sym_cnt       : unsigned (3 downto 0); -- symbol marker index
   signal ms_cnt        : unsigned (3 downto 0); -- millisecond index
   signal pos_marker    : std_logic; -- position marker is on
   signal pos_mark_short: std_logic; -- short position marker
   signal pos_mark_old  : std_logic; -- old position marker
   signal symbol        : std_logic; -- data symbol is on
   signal symbol_short  : std_logic; -- short data symbol
   signal symbol_old    : std_logic; -- old data symbol
   signal p20           : std_logic; -- 20% end marker
   signal p20_old       : std_logic; -- old 20% end marker
   signal p50           : std_logic; -- 50% end marker
   signal p50_old       : std_logic; -- old 50% end marker
   signal p80           : std_logic; -- 80% end marker
   signal p80_old       : std_logic; -- old 80% end marker

begin

   -- short 1 pps
   pps_old <= ONEPPS when rising_edge (CLK);
   pps_short <= '1' when (ONEPPS = '1') and (pps_old = '0') else '0';

   -- generate 1 kHz clock
   MS: process (CLK) is
   begin
      if rising_edge (CLK) then
         kHz <= '0';
         addcycle <= '0';
         if pps_short = '1' then
            cnt <= To_Unsigned (divisor - 4, 24);
            remain <= To_Signed (numerator - denumerator / 2, 16);

         elsif cnt = To_Unsigned (0, 24) then
            if (remain(15) = '1') or (addcycle = '1') then
               kHz <= '1';
               remain <= remain + To_Signed (numerator, 16);
               cnt <= To_Unsigned (divisor - 1, 24);
            else
               addcycle <= '1';
               remain <= remain + To_Signed (-denumerator, 16);
            end if;
         else
            cnt <= cnt - 1;
         end if;
      end if;
   end process MS;

   -- generate position marker index, symbol marker index and millisecond index
   gen_index: process (CLK) is
   begin
      if rising_edge (CLK) then
         if pps_short = '1' then
            pos_cnt <= (others => '0');
            sym_cnt <= (others => '0');
            ms_cnt <= (others => '0');
         elsif kHz = '1' then
            if ms_cnt = To_Unsigned (9, 4) then
               ms_cnt <= (others => '0');
               if sym_cnt = To_Unsigned (9, 4) then
                  sym_cnt <= (others => '0');
                  if pos_cnt = To_Unsigned (9, 4) then
                     pos_cnt <= (others => '0');
                  else
                     pos_cnt <= pos_cnt + 1;
                  end if;
               else
                  sym_cnt <= sym_cnt + 1;
               end if;
            else
               ms_cnt <= ms_cnt + 1;
            end if;
         end if;
      end if;
   end process gen_index;

   -- generate position marker
   gen_pos: process (pos_cnt, sym_cnt) is
   begin
      if (sym_cnt = To_Unsigned (9, 4)) or -- normal markers
         ((sym_cnt = To_Unsigned (0, 4)) and -- additional marker at beginning
          (pos_cnt = To_Unsigned (0, 4))) then
         pos_marker <= '1';
      else
         pos_marker <= '0';
      end if;
   end process gen_pos;
   pos_mark_short <= '1' when (pos_marker = '1') and
                              (ms_cnt = To_Unsigned (0, 4))
                else '0';
   pos_mark_old <= pos_mark_short when rising_edge (CLK);

   -- generate symbol marker
   symbol <= not pos_marker;
   symbol_short <= '1' when (symbol = '1') and (ms_cnt = To_Unsigned (0, 4))
              else '0';
   symbol_old <= symbol_short when rising_edge (CLK);

   -- generate 20%, 50% and 80% marker
   p20 <= '1' when ms_cnt = To_Unsigned (2, 4) else '0';
   p50 <= '1' when ms_cnt = To_Unsigned (5, 4) else '0';
   p80 <= '1' when ms_cnt = To_Unsigned (8, 4) else '0';
   p20_old <= p20 when rising_edge (CLK);
   p50_old <= p50 when rising_edge (CLK);
   p80_old <= p80 when rising_edge (CLK);

   POSMARK <= pos_mark_short and not pos_mark_old;
   SYMMARK <= symbol_short and not symbol_old;
   POSCNT  <= std_logic_vector (pos_cnt);
   SYMCNT  <= std_logic_vector (sym_cnt);
   POSEND  <= pos_marker and p80 and not p80_old;
   ZEROEND <= symbol and p20 and not p20_old;
   ONEEND  <= symbol and p50 and not p50_old;

end Structure;
--------------------------------------------------------------------------------
