--------------------------------------------------------------------------------
-- SubModule IRIGB_decoder
-- Created   2/2/2009 2:30:29 PM
--
-- CLK needs to be 2^26 Hz. Otherwise, the constants need to be changed
-- accordingly.
--
-- BITS encodes all data bits of an IRIG-B signal ( 10 * 9 - 1).
-- The LSB is bit 0 of the second count
-- BITS( 3 downto  0): lower digit of second (BCD)
-- BITS( 4)          : filler
-- BITS( 7 downto  5): upper digit of second (BCD)
-- BITS(11 downto  8): lower digit of minute (BCD)
-- BITS(12)          : filler
-- BITS(15 downto 13): upper digit of minute (BCD)
-- BITS(16)          : filler
-- BITS(20 downto 17): lower digit of hour (BCD)
-- BITS(21)          : filler
-- BITS(23 downto 22): upper digit of hour (BCD)
-- BITS(25 downto 24): filler
-- BITS(29 downto 26): lower digit of day (BCD)
-- BITS(30)          : filler
-- BITS(34 downto 31): middle digit of day (BCD)
-- BITS(36 downto 35): upper digit of hour (BCD)
-- BITS(43 downto 37): filler
-- BITS(70 downto 44): control functions
-- BITS(87 downto 71): straight binary seconds (17 bits)
-- BITS(88)          : filler
--
-- With IEEE 1344:
-- BITS (47 downto 44): lower digit of year (BCD)
-- BITS (48)          : filler
-- BITS (52 downto 49): upper digit of year (BCD)
-- BITS (53)          : leap second pending
-- BITS (54)          : leap second: 0-add, 1-subtract
-- BITS (55)          : daylight saving pending
-- BITS (56)          : daylight savings time
-- BITS (57)          : time offset sign
-- BITS (61 downto 58): timezone hour offset (binary, 4 bits)
-- BITS (62)          : time offset-0.5 hour
-- BITS (66 downto 63): time quality
-- BITS (67)          : parity
-- BITS (70 downto 69): not used
--
-- When a valid IRIG-B signal has been detected VALID is asserted for 1
-- clock cycle. INVALID is asserted for one clock cycle when an illegale
-- IRIG-B signal has been encountered and once a second if no valid start
-- marker can be detected.
--
-- ONEPPS goes high for one clock cycle with the start of a new IRIG-B
-- frame. It is synchronized with CLK.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity IRIGB_decoder is port
   (
     RST            : in    std_logic;
     CLK            : in    std_logic;
     IRIG           : in    std_logic;
     ONEPPS         : out   std_logic;
     VALID          : out   std_logic;
     INVALID        : out   std_logic;
     BITS           : out   std_logic_vector (88 downto 0)
   );
end IRIGB_decoder;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of IRIGB_decoder is

-- Constants
   constant divisor     : integer := 671089; -- 2**26 / 100;
   constant threshold5  : integer := divisor *  5 / 100;
   constant threshold35 : integer := divisor * 35 / 100;
   constant threshold65 : integer := divisor * 65 / 100;
   constant threshold95 : integer := divisor * 95 / 100;

-- Type Declarations
   type pulse_type is (UNDER, SHORT, MEDIUM, LONG, OVER);
   type state_type is (IDLE, FRAME, OK, BAD);

-- Signal Declarations
   signal pulse         : pulse_type;
   signal prev_pulse    : pulse_type;
   signal up            : std_logic;
   signal down          : std_logic;
   signal irig_sample   : std_logic_vector (1 downto 0);
   signal count         : unsigned (19 downto 0);
   signal shift         : std_logic_vector (88 downto 0);
   signal state         : state_type;
   signal position      : unsigned (3 downto 0);
   signal bitcnt        : unsigned (3 downto 0);
   signal timeout       : unsigned (25 downto 0);
   signal onepps_enable : std_logic;

begin

   -- sample irig signal
   irig_sampling: process (CLK) is
   begin
      if rising_edge (CLK) then
         irig_sample (0) <= irig_sample (1);
         irig_sample (1) <= irig;
      end if;
   end process irig_sampling;
   up   <=     irig_sample (1) and not irig_sample (0);
   down <= not irig_sample (1) and     irig_sample (0);

   -- determine pulse type
   pulse_width: process (CLK) is
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            pulse <= OVER;
         elsif up = '1' then
            count <= To_unsigned (0, 20);
            pulse <= UNDER;
            prev_pulse <= pulse;
         elsif down = '1' then
            count <= To_Unsigned (threshold95 + 1, 20);
         else
            if count = To_Unsigned (threshold5, 20) then
               pulse <= SHORT;
            elsif count = To_Unsigned (threshold35, 20) then
               pulse <= MEDIUM;
            elsif count = To_Unsigned (threshold65, 20) then
               pulse <= LONG;
            elsif count = To_Unsigned (threshold95, 20) then
               pulse <= OVER;
            end if;
            if count /= To_Unsigned (threshold95 + 1, 20) then
               count <= count + 1;
            end if;
         end if;
      end if;
   end process pulse_width;

   -- process data
   data_processing: process (RST, CLK) is
   begin
      if RST = '1' then 
         BITS <= (others => '0');
         VALID <= '0';
         INVALID <= '0';
         position <= To_Unsigned (0, 4);
         bitcnt <= To_Unsigned (0, 4);
         timeout <= (others => '0');
         shift <= (others => '0');
         state <= IDLE;
      elsif rising_edge (CLK) then
         case state is
            -- wait for start pulse
            when IDLE =>
               VALID <= '0';
               INVALID <= '0';
               timeout <= timeout + 1;
               if down = '1' then
                  -- check for start marker P_r
                  if (prev_pulse = LONG) and (pulse = LONG) then
                     position <= To_Unsigned (0, 4);
                     bitcnt <= To_Unsigned (1, 4);
                     timeout <= (others => '0');
                     state <= FRAME;
                  end if;
               end if;
               -- timeout is reached every 1 second when no P_r is detected
               if timeout = B"11" & X"FFFFFF" then
                  BITS <= (others => '0');
                  INVALID <= '1';
               end if;
            -- frame processing
            when FRAME =>
               VALID <= '0';
               INVALID <= '0';
               -- timeout is 1 second (should finish in 990ms!)
               if timeout = B"11" & X"FFFFFF" then
                  state <= BAD;
               end if;
               timeout <= timeout + 1;
               -- a bit has been detected
               if down = '1' then
                  -- is this the position marker following the frame?
                  if bitcnt = To_Unsigned (9, 4) then
                     if PULSE /= LONG then
                        -- error if not a position marker
                        state <= BAD;
                     else
                        -- check if this is the last position marker P_0
                        if position = To_Unsigned (9, 4) then
                           -- last bit processed
                           state <= OK;
                        else
                           -- move to next frame
                           position <= position + 1;
                           bitcnt <= To_Unsigned (0, 4);
                        end if;
                     end if;
                  -- data bit expected
                  else
                     if PULSE = SHORT then
                        -- short pulse is zero
                        shift <= '0' & shift (88 downto 1);
                        bitcnt <= bitcnt + 1;
                     elsif pulse = MEDIUM then
                        -- medium pulse is one
                        shift <= '1' & shift (88 downto 1);
                        bitcnt <= bitcnt + 1;
                     else
                        -- all others are illegale
                        state <= BAD;
                     end if;
                  end if;
               end if;
            -- a valid IRIG signal has been encoded
            when OK =>
               BITS <= shift;
               timeout <= (others => '0');
               VALID <= '1';
               state <= IDLE;
            -- a illegale IRIG signal has been encountered
            when BAD =>
               BITS <= (others => '0');
               timeout <= (others => '0');
               INVALID <= '1';
               state <= IDLE;
         end case;
      end if;
   end process data_processing;

   -- 1 PPS detection
   pps: process (CLK) is
   begin
      if rising_edge (CLK) then
         if state = OK then
            -- next positive edge is 1 second mark
            onepps_enable <= '1';
         elsif down = '1' then
            onepps_enable <= '0';
         end if;
      end if;
   end process pps;
   ONEPPS <= up and onepps_enable;

end architecture Structure;


