--------------------------------------------------------------------------------
-- SubModule IRIGB_encoder
-- Created   2/2/2009 2:30:29 PM
--
-- The maximum number of cycles needed to encode the IRIG B signal from
-- a GPS time is 132. Since the data processing is started with the 1PPS
-- and is registered at the 80% mark of the first symbol (position marker Pr),
-- the minimum clock frequency on CLK is 16.5kHz.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity IRIGB_encoder is port
   (
     CLK            : in    std_logic;
     ONEPPS         : in    std_logic;
     GPS            : in    std_logic_vector(31 downto 0);
     BITS           : out   std_logic_vector(88 downto 0);
     LEAP           : in    std_logic_vector(7 downto 0);
     LEAPPEND       : in    std_logic;
     LEAPDEL        : in    std_logic;
     ISLEAP         : out   std_logic;
     ISLAST         : out   std_logic;
     TIMEZONE       : in    std_logic_vector(5 downto 0);
     DST            : in    std_logic;
     IEEE1344       : in    std_logic
   );
end IRIGB_encoder;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of IRIGB_encoder is

-- Constant Declarations
   constant TZ_SEC       : integer := 1800;
   constant DAY100_SEC   : integer := 8640000;
   constant DAY10_SEC    : integer := 864000;
   constant DAY1_SEC     : integer := 86400;
   constant HOUR10_SEC   : integer := 36000;
   constant HOUR1_SEC    : integer := 3600;
   constant MINUTE10_SEC : integer := 600;
   constant MINUTE1_SEC  : integer := 60;
   constant SECOND10_SEC : integer := 10;
   -- Possible times in the year for adding/subtracting leap seconds
   -- Calculate backwards from the next year (avoids leap year calculation!)
   constant leap4        : integer :=                -1; -- end of Dec
   constant leap3        : integer := - 92*DAY1_SEC - 1; -- end of Sep
   constant leap2        : integer := -184*DAY1_SEC - 1; -- end of Jun
   constant leap1        : integer := -275*DAY1_SEC - 1; -- end of Mar

-- Component Declarations (same ROM as in TimeDecoder)
   component ROM_1024_32
       port (
       clka:  IN  std_logic;
       addra: IN  std_logic_vector (9 downto 0);
       douta: OUT std_logic_vector (31 downto 0));
   end component;

-- Type Declarations
   type getyear_type is (IDLE, PREP1, PREP2, PREP3A, PREP3B, PREP4, PREP5,
                         YEAR1, DAY100, DAY10, DAY1, DAY1B, DAY1C,
                         HOUR10, HOUR1, MINUTE10, MINUTE1,
                         SECOND10, FINISH);

-- Signal Declarations
   signal pps_old       : std_logic;  -- old 1 PPS
   signal pps_oldold    : std_logic;  -- old old 1 PPS
   signal pps_short     : std_logic;  -- a short version of the 1 PPS
   signal addr          : std_logic_vector (9 downto 0); -- lookup address
   signal lookup        : std_logic_vector (31 downto 0); -- lookup GPS value
   signal tz            : signed (5 downto 0); -- time zones in 30 min
   signal year          : unsigned (5 downto 0); -- year since 2000
   signal is_leap_year  : std_logic; -- is a leap year
   signal is_leap       : std_logic; -- is a leap second
   signal add_leap      : std_logic; -- is a +leap second
   signal sub_leap      : std_logic; -- is a -leap second (one before!)
   signal is_last       : std_logic; -- is last second of year
   signal add_leap4     : std_logic; -- is +leap second at end of Dec
   signal add_leap3     : std_logic; -- is +leap second at end of Sep
   signal add_leap2     : std_logic; -- is +leap second at end of Jun
   signal add_leap1     : std_logic; -- is +leap second at end of Mar
   signal sub_leap4     : std_logic; -- is -leap second at end of Dec
   signal sub_leap3     : std_logic; -- is -leap second at end of Sep
   signal sub_leap2     : std_logic; -- is -leap second at end of Jun
   signal sub_leap1     : std_logic; -- is -leap second at end of Mar
   signal mark_leap     : std_logic; -- mark pending leap second
   signal y10           : unsigned (3 downto 0); -- year tens
   signal y1            : unsigned (3 downto 0); -- year last digits
   signal d100          : unsigned (3 downto 0); -- day first digit
   signal d10           : unsigned (3 downto 0); -- day middle digit
   signal d1            : unsigned (3 downto 0); -- day last digit
   signal h10           : unsigned (3 downto 0); -- hour first digit
   signal h1            : unsigned (3 downto 0); -- hour last digit
   signal m10           : unsigned (3 downto 0); -- minute first digit
   signal m1            : unsigned (3 downto 0); -- minute last digit
   signal s10           : unsigned (3 downto 0); -- seconds first digit
   signal s1            : unsigned (3 downto 0); -- seconds last digit
   signal sbs           : unsigned (16 downto 0);-- straight binary sec
   signal diff          : signed (32 downto 0);  -- difference
   signal sdiff         : signed (32 downto 0);  -- difference
   signal old_diff      : signed (32 downto 0);  -- old difference value
   signal next_diff     : signed (32 downto 0);  -- difference after -year
   signal state         : getyear_type;          -- state of decoder
   signal data          : unsigned(88 downto 0); -- data bits
   signal parity        : std_logic;             -- parity of data bits

begin
   -- short 1 pps
   pps_old <= ONEPPS when rising_edge (CLK);
   pps_oldold <= pps_old when rising_edge (CLK);
   pps_short <= '1' when (pps_old = '1') and (pps_oldold = '0') else '0';

   -- Lookup ROM
   lookup_ROM : ROM_1024_32
        port map (
            clka => CLK,
            addra => addr,
            douta => lookup);
   addr <= std_logic_vector (year) & B"0001";
   next_diff <= diff - (B"0" & signed (lookup)) when rising_edge (CLK);
   -- check for leap year (works till 2100, beyond my retirement!)
   is_leap_year <= '1' when (year(1 downto 0) = B"00") else '0';

   -- check for leap seconds at end of Dec, Sep, Jun or Mar
   add_leap4 <= '1' when ((next_diff = To_Signed (leap4+1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '0'))
                    else '0';
   sub_leap4 <= '1' when ((next_diff = To_Signed (leap4-1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '1'))
                    else '0';
   add_leap3 <= '1' when ((next_diff = To_Signed (leap3+1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '0'))
                    else '0';
   sub_leap3 <= '1' when ((next_diff = To_Signed (leap3-1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '1'))
                    else '0';
   add_leap2 <= '1' when ((next_diff = To_Signed (leap2+1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '0'))
                    else '0';
   sub_leap2 <= '1' when ((next_diff = To_Signed (leap2-1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '1'))
                    else '0';
   add_leap1 <= '1' when ((next_diff = To_Signed (leap1+1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '0'))
                    else '0';
   sub_leap1 <= '1' when ((next_diff = To_Signed (leap1-1, 33)) and
                          (LEAPPEND = '1') and (LEAPDEL = '1'))
                    else '0';
   add_leap <= add_leap1 or add_leap2 or add_leap3 or add_leap4;
   sub_leap <= sub_leap1 or sub_leap2 or sub_leap3 or sub_leap4;
   is_leap <= add_leap or sub_leap;
   mark_leap <= is_leap;
   ISLEAP <= is_leap;

   --is_leap <= '1' when (next_diff = To_Signed (0, 33)) and (LEAPPEND = '1')
         --else '0'; -- only 1 when a leap second is added!
   --mark_leap <= '1' when ((LEAPPEND = '1') and
   --                       (next_diff (32 downto 5) = X"FFFFFFF")) or
   --                      (is_leap = '1')
   --        else '0'; -- mark pending leap sec dutring the last 32 seconds

   -- check for last second in year
   is_last <= '1' when ((next_diff = To_Signed (-1, 33)) and
                       (LEAPPEND = '0')) or
                       (add_leap4 = '1') or (sub_leap4 = '1')
                  else '0';
   ISLAST <= is_last;

   -- get the year information
   getyeardaytime: process (CLK) is
   begin
      if rising_edge (CLK) then
         if pps_short = '1' then
            year <= (others => '0');
            y10 <= To_Unsigned (9, 4);
            y1  <= To_Unsigned (9, 4);
            d100 <= To_Unsigned (15, 4);
            d10  <= To_Unsigned (15, 4);
            d1   <= To_Unsigned (15, 4);
            h10  <= To_Unsigned (15, 4);
            h1   <= To_Unsigned (15, 4);
            m10  <= To_Unsigned (15, 4);
            m1   <= To_Unsigned (15, 4);
            s10  <= To_Unsigned (15, 4);
            s1   <= To_Unsigned (0, 4);
            old_diff <= (others => '0');
            tz <= signed (TIMEZONE);
            state <= PREP1;
         else case state is
            -- Preparation
            when PREP1 =>
               diff <= B"0" & signed (GPS);
               state <= PREP2;
            -- adjust for leap seconds
            when PREP2 =>
               diff <= diff - (B"0" & X"000000" & signed (LEAP));
               if tz(5) = '1' then
                  state <= PREP3B;
               else
                  state <= PREP3A;
               end if;
            -- adjust for positive timezones
            when PREP3A =>
               if tz = To_Signed (0, 6) then
                  state <= PREP4;
               else
                  tz <= tz - 1;
                  diff <= diff - To_Signed (-TZ_SEC, 33);
               end if;
            -- adjust for negative timezones
            when PREP3B =>
               if tz = To_signed (0, 6) then
                  state <= PREP4;
               else
                  tz <= tz + 1;
                  diff <= diff - To_Signed (TZ_SEC, 33);
               end if;
            -- adjust for daylight savings
            when PREP4 =>
               if DST = '1' then
                  diff <= diff - To_Signed (-HOUR1_SEC, 33);
               end if;
               state <= PREP5;
               year <= year + 1;
            -- delay for calc next_diff
            when PREP5 =>
               state <= YEAR1;
            -- compute year
            when YEAR1 =>
               old_diff <= next_diff;
               if add_leap = '1' then
                  year <= year - 1; -- readjust
                  sdiff <= old_diff - 1; -- treat it the same as last second
                                         -- of the month, later add it again
                  state <= DAY100;
               elsif next_diff(32) = '1' then
                  year <= year - 1; -- readjust
                  sdiff <= old_diff;
                  state <= DAY100;
               elsif year = B"000000" then
                  year <= year - 1; -- readjust
                  sdiff <= (others => '0');
                  state <= DAY100;
               else
                  year <= year + 1;
                  if y1 = To_Unsigned (9, 4) then
                     y1 <= To_Unsigned (0, 4);
                     if y10 = To_Unsigned (9, 4) then
                        y10 <= To_Unsigned (0, 4);
                     else
                        y10 <= y10 + 1;
                     end if;
                  else
                     y1 <= y1 + 1;
                  end if;
                  state <= PREP5;
               end if;
            -- compute 100 days
            when DAY100 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (d100 = To_Unsigned (9, 4)) then
                  sdiff <= old_diff;
                  state <= DAY10;
               else
                  sdiff <= sdiff - To_signed (DAY100_SEC, 33);
                  d100 <= d100 + 1;
               end if;
            -- compute 10 days
            when DAY10 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (d10 = To_Unsigned (9, 4)) then
                  sdiff <= old_diff;
                  state <= DAY1;
               else
                  sdiff <= sdiff - To_signed (DAY10_SEC, 33);
                  d10 <= d10 + 1;
               end if;
            -- compute 1 days
            when DAY1 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (d1 = To_Unsigned (9, 4)) then
                  sdiff <= old_diff;
                  sbs <= unsigned (old_diff (16 downto 0));
                  state <= DAY1B;
                  d1 <= d1 + 1; -- add one!
               else
                  sdiff <= sdiff - To_signed (DAY1_SEC, 33);
                  d1 <= d1 + 1;
               end if;
            -- Days start with 1 on Jan 01
            when DAY1B =>
               if d1 = To_Unsigned (10, 4) then
                  d1 <= To_Unsigned (0, 4);
                  d10 <= d10 + 1;
                  state <= DAY1C;
               else
                   state <= HOUR10;
               end if;
            when DAY1C =>
               if d10 = To_Unsigned (10, 4) then
                  d10 <= To_Unsigned (0, 4);
                  d100 <= d100 + 1;
               end if;
               state <= HOUR10;
            -- Hour10
            when HOUR10 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (h10 = To_Unsigned (9, 4)) then
                  sdiff <= old_diff;
                  state <= HOUR1;
               else
                  sdiff <= sdiff - To_signed (HOUR10_SEC, 33);
                  h10 <= h10 + 1;
               end if;
            -- Hour1
            when Hour1 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (h1 = To_Unsigned (9, 4)) then
                  sdiff <= old_diff;
                  state <= MINUTE10;
               else
                  sdiff <= sdiff - To_signed (HOUR1_SEC, 33);
                  h1 <= h1 + 1;
               end if;
            -- Minute10
            when MINUTE10 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (m10 = To_Unsigned (9, 4)) then
                  sdiff <= old_diff;
                  state <= MINUTE1;
               else
                  sdiff <= sdiff - To_signed (MINUTE10_SEC, 33);
                  m10 <= m10 + 1;
               end if;
            -- Minute1
            when MINUTE1 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (m1 = To_Unsigned (9, 4)) then
                  if add_leap = '1' then
                     sdiff <= To_signed (60, 33); -- add leap second again
                  else 
                     sdiff <= old_diff;
                  end if;  
                  state <= SECOND10;
               else
                  sdiff <= sdiff - To_signed (MINUTE1_SEC, 33);
                  m1 <= m1 + 1;
               end if;
            -- Seconds10
            when SECOND10 =>
               old_diff <= sdiff;
               if (sdiff(32) = '1') or (s10 = To_Unsigned (9, 4)) then
                  sdiff <= old_diff;
                  s1 <= unsigned (old_diff (3 downto 0));
                  state <= FINISH;
               else
                  sdiff <= sdiff - To_signed (SECOND10_SEC, 33);
                  s10 <= s10 + 1;
               end if;
            -- finish
            when FINISH =>
               BITS <= std_logic_vector (data);
            --- idle
            when IDLE =>
            end case;
         end if;
      end if;
   end process getyeardaytime;

   -- assign data bits
   data (3 downto 0) <= s1;
   data (4) <= '0';
   data (7 downto 5) <= s10 (2 downto 0);
   data (11 downto 8) <= m1;
   data (12) <= '0';
   data (16 downto 13) <= m10;
   data (20 downto 17) <= h1;
   data (21) <= '0';
   data (25 downto 22) <= h10;
   data (29 downto 26) <= d1;
   data (30) <= '0';
   data (34 downto 31) <= d10;
   data (38 downto 35) <= d100;
   data (43 downto 39) <= (others => '0');
   data (87 downto 71) <= sbs;
   data (88) <= '0';
   process (IEEE1344, y1, y10, mark_leap, LEAPDEL, DST, TIMEZONE, parity) is
   begin
      if IEEE1344 = '1' then
         data (47 downto 44) <= y1;
         data (48) <= '0';
         data (52 downto 49) <= y10;
         data (53) <= mark_leap; -- leap sec pending
         data (54) <= mark_leap and LEAPDEL;   -- delete a second
         data (55) <= '0';       -- no announcement of DST
         data (56) <= DST;       -- DST flag
         data (62 downto 57) <= unsigned (TIMEZONE);
         data (66 downto 63) <= B"0100"; -- 1 us accuracy
         data (67) <= parity;
         data (70 downto 68) <= B"000";
      else
         data (70 downto 44) <= (others => '0');
      end if;
   end process;
   parity <= data(0)  xor data(1)  xor data(2)  xor data(3)  xor data (4)  xor
             data(5)  xor data(6)  xor data(7)  xor data(8)  xor data (9)  xor
             data(10) xor data(11) xor data(12) xor data(13) xor data (14) xor
             data(15) xor data(16) xor data(17) xor data(18) xor data (19) xor
             data(20) xor data(21) xor data(22) xor data(23) xor data (24) xor
             data(25) xor data(26) xor data(27) xor data(28) xor data (29) xor
             data(30) xor data(31) xor data(32) xor data(33) xor data (34) xor
             data(35) xor data(36) xor data(37) xor data(38) xor data (39) xor
             data(40) xor data(41) xor data(42) xor data(43) xor data (44) xor
             data(45) xor data(46) xor data(47) xor data(48) xor data (49) xor
             data(50) xor data(51) xor data(52) xor data(53) xor data (54) xor
             data(55) xor data(56) xor data(57) xor data(58) xor data (59) xor
             data(60) xor data(61) xor data(62) xor data(63) xor data (64) xor
             data(65) xor data(66);

end Structure;
--------------------------------------------------------------------------------
