--------------------------------------------------------------------------------
-- OnePPSCounter: Counts the number of ticks between two 1PPS events
-- LIGO document: D070000-00
-- SubModule OnePPSCounter
--   The two trigger inputs A and B are meant to be driven by two 1PPS signals.
--   This module always measures the short interval and, therefore,
--      all measured output values will be within -0.5 <= t < +0.5.
--   A positive t corresponds to A triggering before B, whereas a negative t
--      corresponds to B triggering before A.
--   This module uses the negative clock edge of CLK to deduce a half count.
--   The timing resolution is approximately +/- 2^-27/sqrt(12) seconds with
--      a 2^26Hz clock input (approx. +/-2.2ns).
--   If the 1PPS on input A is synchronous with the clock, the reported
--      interval will be biased by -2^-28s (approx. -3.7ns)
-- 1PPS inputs (positive edge)
--   INA: First input
--   INB: Second input
-- The clock frequency is set to 2^26 Hz  (2^counter_width)
--   CLK: basic clock
-- Output:
--   COUNT(31 downto 0): time interval in units of 2^-32 sec
--      Interpret as 2's complement, ie., the MSB is the sign
--      The time resolution is 2^-27; the trailing bits are filled with zeros.
--   VALID: Goes high for one clock cycle when a new output time interval
--      value is ready at COUNT.
-- Created   7/29/2007
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;
--synopsys translate_off
--library UNISIM;
--use unisim.vcomponents.all;
--synopsys translate_on

entity OnePPSCounter is port
   (
   CLK               : in  std_logic;
   INA               : in  std_logic;
   INB               : in  std_logic;
   COUNT             : out std_logic_vector (31 downto 0);
   VALID             : out std_logic
   );
end entity OnePPSCounter;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of OnePPSCounter is

-- Components

-- Const Declarations
   constant counter_width  : integer := 26;
   constant counter_zeros  : integer := 32 - counter_width - 1;
   constant timeout_zero   : unsigned (counter_width-1 downto 0)
                                := (others => '0');

-- Type Declarations

-- Signal Declarations
   signal counter        : unsigned (counter_width-1 downto 0):=(others => '0');
   signal ina_old        : std_logic;
   signal ina_oldold     : std_logic;
   signal ina_half       : std_logic;
   signal ina_prehalf    : std_logic;
   signal inb_old        : std_logic;
   signal inb_oldold     : std_logic;
   signal inb_half       : std_logic;
   signal inb_prehalf    : std_logic;
   signal triga          : std_logic;
   signal triga_old      : std_logic;
   signal trigb          : std_logic;
   signal trigb_old      : std_logic;
   signal marka          : unsigned (counter_width downto 0);
   signal markb          : unsigned (counter_width downto 0);
   signal diff           : unsigned (counter_width downto 0);
   signal timeout        : unsigned (counter_width-1 downto 0);
   signal timeout_tc     : std_logic;

begin
   -- counter
   counter <= counter - 1 when rising_edge (CLK);

   -- input triggers
   ina_old <= ina when rising_edge (CLK);
   ina_oldold <= ina_old when rising_edge (CLK);
   ina_prehalf <= ina when falling_edge (CLK);
   ina_half <= ina_prehalf when rising_edge (CLK);
   inb_old <= inb when rising_edge (CLK);
   inb_oldold <= inb_old when rising_edge (CLK);
   inb_prehalf <= inb when falling_edge (CLK);
   inb_half <= inb_prehalf when rising_edge (CLK);
   triga <= '1' when (ina_oldold = '0') and (ina_old = '1') else '0';
   trigb <= '1' when (inb_oldold = '0') and (inb_old = '1') else '0';
   triga_old <= triga when rising_edge (CLK);
   trigb_old <= trigb when rising_edge (CLK);

   -- time stamp when input a goes high
   trigger_a: process (CLK, triga, ina_half) is
   begin
      if rising_edge (CLK) then
         if triga = '1' then
            marka (counter_width downto 1) <= counter;
            marka (0) <= ina_half;
         end if;
      end if;
   end process trigger_a;

   -- time stamp when input b goes high
   trigger_b: process (CLK, trigb, inb_half) is
   begin
      if rising_edge (CLK) then
         if trigb = '1' then
            markb (counter_width downto 1) <= counter;
            markb (0) <= inb_half;
         end if;
      end if;
   end process trigger_b;

   -- compute difference between timestamps
   diff <= marka - markb;

   -- check if we have a valid result
   is_valid: process (CLK, triga_old, trigb_old, diff, timeout_tc) is
   begin
      if rising_edge (CLK) then
         if (triga_old = '1') and (trigb_old = '1') then
            COUNT (31 downto counter_zeros) <= std_logic_vector (diff);
            VALID <= '1';
         elsif (triga_old = '1') and (trigb_old = '0') then
            if (diff (counter_width) = '1') and (timeout_tc = '0') then
               COUNT (31 downto counter_zeros) <= std_logic_vector (diff);
               VALID <= '1';
            else
               VALID <= '0';
            end if;
         elsif (triga_old = '0') and (trigb_old = '1') then
            if (diff (counter_width) = '0') and (timeout_tc = '0') then
               COUNT (31 downto counter_zeros) <= std_logic_vector (diff);
               VALID <= '1';
            else
               VALID <= '0';
            end if;
         else
            VALID <= '0';
         end if;
      end if;
   end process is_valid;

   COUNT (counter_zeros - 1 downto 0) <= (others => '0');

   -- timeout is 0.56 sec
   do_timeout: process (CLK, triga_old, trigb_old, timeout_tc) is
   begin
      if rising_edge (CLK) then
         if (triga_old = '1') or (trigb_old = '1') then
            timeout (counter_width-1 downto counter_width-4) <= B"1001";
            timeout (counter_width-5 downto 0) <= (others => '0');
         elsif timeout_tc = '0' then
            timeout <= timeout - 1;
         end if;
      end if;
   end process do_timeout;
   timeout_tc <= '1' when timeout = timeout_zero else '0';

end architecture Structure;

