--------------------------------------------------------------------------------
-- EEPROM Interface
--
-- Ilya Belopolski
-- 19 August, 2010
--------------------------------------------------------------------------------
library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity RunEEPROM is port
   (
     CLK           : in    std_logic;                         -- The clock, 2^26 Hz = 67MHz

     -- EEPROM-side interface
     CFG_SI        : in    std_logic;                         -- Signal sent out by the EEPROM
     CFG_CS        : out   std_logic;                         -- Activation signal for the EEPROM
     CFG_SO        : out   std_logic;                         -- Signal sent into the EEPROM
     CFG_CLK       : out   std_logic;                         -- Clock for the EEPROM, maximum 5MHz at 1.8V operation.

     -- FPGA-side interface
     RST           : in    std_logic;                         -- Prepares the EEPROM for a new command
     REI           : in    std_logic;                         -- Reinitializes the EEPROM, i.e. sends the start command & sets the status buffer.
     RDY           : out   std_logic;                         -- Indicates whether the given command has been completed
     RDWR          : in    std_logic;                         -- Toggles read & write mode
     ADDR          : in    std_logic_vector (15 downto 0);    -- An address on the EEPROM
     WDATA         : in    std_logic_vector (7 downto 0);     -- Data sent to a given address on the EEPROM
     RDATA         : out   std_logic_vector (7 downto 0);     -- Data read from a given address on the EEPROM
     SYSRST        : in    std_logic

   );
end RunEEPROM;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture ark of RunEEPROM is

     -- The message refers to what we hope to accomplish, e.g. read some data off the chip
     type mess_type is (INIT, RD, WR, RDST) ; -- READ seemed too special...
     signal mess : mess_type := INIT;

     -- The state refers to where we are in the current message: sending the command, listening for output, etc.
     type state_type is (START, EEND, SEND, RECEIVE) ; -- END was taken...
     signal state : state_type := START;

--     signal ini : boolean := false; -- Makes sure that the chip is initialized at start-up
     -- (Initialization means: send the WREN command and use the WRSR command to set the status buffer)

     signal rdyrdy : std_logic := '0'; -- Nothing but a buffer for the RDY signal

     -- Repository of commands
     constant WREN : std_logic_vector := B"0000_0110"; -- An initialization command, gets the chip listening
     constant WRSR : std_logic_vector := B"0000_0001"; -- Another initialization command...
     constant STAT : std_logic_vector := B"0000_0010"; -- The status vector we want to load in (follows the WRSR command);
     constant REED : std_logic_vector := B"0000_0011"; -- The command to read from memory
     constant WRIT : std_logic_vector := B"0000_0010"; -- The command to write to memory

     -- The EEPROM signals
     signal cs     : std_logic;
     signal sclk   : std_logic;
     signal si     : std_logic_vector (7 downto 0);
     signal so     : std_logic;

     -- The clock generator module
     component EEPROMClock is port
     (
          CLK     : in  std_logic;
          HALFT   : in  std_logic_vector (31 downto 0);
          divCLK  : out std_logic
     );
     end component EEPROMClock;

     -- Timing signals for the clock
     signal divCLK         : std_logic;
     signal divCLK_old     : std_logic;
     signal divCLK_ancient : std_logic;
     signal rise           : std_logic;
     signal fall           : std_logic;

     -- Turns the clock on and off...
     signal tickCLK        : boolean := false;

     -- Some useful flags for... flagging
     signal flag           : boolean := false;
     signal sendCMD        : boolean := true;
     signal sendDAT        : boolean := false;

     -- The bit index, for sending bits in one after another
     signal j              : integer := 0;

     -- Keep a short history of RST
     signal oldRST         : std_logic;

     signal delay          : integer := 0;

begin

     -- Set up the clock generator module
     makeClock : EEPROMClock
     port map (
          CLK => CLK,            -- The (very, very fast) FPGA CLK
          HALFT => X"0000_1000", -- A more or less random choice for the clock period
          -- Note that the EEPROM chip operate between 2.5V & GND, which
          -- restricts the clock speed to maximum of 5MHz... but speed isn't
          -- important for us, so we clock well below this limit.
          divCLK => divCLK       -- The (rather slow) EEPROM CLK
     );

     divCLK_old <= divCLK when rising_edge (CLK);
     divCLK_ancient <= divCLK_old when rising_edge (CLK);
     rise <= '1' when (divCLK_ancient = '0') and (divCLK_old = '1') else '0';
     fall <= '1' when (divCLK_ancient = '1') and (divCLK_old = '0') else '0';

     oldRST <= RST when rising_edge (CLK);

     RDY <= rdyrdy;

     CFG_CS <= cs;
     CFG_CLK <= sclk;
     CFG_SO <= so;

     RDATA <= si;

     talk : process (CLK) is
     begin
          if rising_edge (CLK) then

             if (SYSRST = '1') then

                mess <= INIT;
                state <= START;
--             ini <= false;
                rdyrdy <= '0';

                tickCLK <= false;
                flag <= false;
                sendCMD <= true;
                sendDAT <= false;
                j <= 0;
                delay <= 0;
                cs <= '1';

             else

             -- Based on input, determine the command --

     -- Trigger on the rising edge of RST.
     -- If triggered, set the state to START and RDY <= '0';
     -- When finished, stay in EEND until called back to start by this loop.

     -- Note that we ignore commands if RDY = '0'.

--             if (not ini) then
--                ini <= true;
--                mess <= INIT;
--                rdyrdy <= '0';
--                state <= START;
--             end if;

             if (oldRST = '0') and (RST = '1') and (rdyrdy = '1') then
                if (REI = '1') then
                    mess <= INIT;
                    rdyrdy <= '0';
                    flag <= false;
                    state <= START;
                elsif (RDWR = '0') then
                    mess <= RD;
                    rdyrdy <= '0';
                    flag <= false;
                    state <= START;
                elsif (RDWR = '1') then
                    mess <= WR;
                    rdyrdy <= '0';
                    flag <= false;
                    state <= START;
                end if;
             end if;

             -- The chip clock --
             if tickCLK then
                if (rise = '1') then
                   sclk <= '1';
                elsif (fall = '1') then
                   sclk <= '0';
                end if;
             end if;

             -- Talk to the chip --
             if (state = START) then -- Bring CS low, wait a while...
                cs <= '0';
                sclk <= '0';
                if (rise = '1') and (flag) then
                   flag <= false;
                   state <= SEND;
                elsif (rise = '1') then
                   flag <= true;
                end if;
             elsif (state = EEND) then
                tickCLK <= false;
                cs <= '1';
                if (fall = '1') and (flag) then
                   if (not sendCMD) then
                      state <= START;
                      flag <= false;
--                   elsif (mess = INIT) and (delay < 0) then
                      -- Pause for a bit --
--                      delay <= delay + 1;
                   else
                      rdyrdy <= '1';
--                      delay <= 0;
                   end if;
                elsif (fall = '1') then
                   flag <= true;
                end if;
             elsif (state = SEND) then
                if (fall = '1') then
                   if (mess = INIT) then
                   -- For INIT, send three messages: WREN, WRSR, and a status vector.

                      if (sendCMD) then
                         so <= WREN(j);
                         j <= j + 1;
                         tickCLK <= true;
                         if (j > 7) then
                            j <= 0;
                            sendCMD <= false;
                            state <= EEND;
                         end if;
                      elsif (not sendCMD) and (not sendDAT) then
                         so <= WRSR(j);
                         j <= j + 1;
                         tickCLK <= true;
                         if (j = 7) then
                            j <= 0;
                            sendDAT <= true;
                         end if;
                      elsif (not sendCMD) and (sendDAT) then
                         so <= STAT(j);
                         j <= j + 1;
                         if (j > 7) then
                            j <= 0;
                            sendCMD <= true;
                            sendDAT <= false;
                            state <= EEND;
                         end if;
                      end if;

                   elsif (mess = RD) then
                   -- For REED, send the read instruction & an address (note: the address is 16 bits)

                      if (sendCMD) then
                         so <= REED(j);
                         j <= j + 1;
                         tickCLK <= true;
                         if (j = 7) then
                            j <= 0;
                            sendCMD <= false;
                         end if;
                      elsif (not sendCMD) then
                         so <= ADDR(j);
                         j <= j + 1;
                         if (j > 15) then
                            j <= 0;
                            sendCMD <= true;
                            state <= RECEIVE;
                         end if;
                      end if;

                   elsif (mess = WR) then
                   -- For WRIT, send WREN, the write instruction, an address (16 bits), & data (8 bits);

                      if (sendCMD) and (not sendDAT) then
                         so <= WREN(j);
                         j <= j + 1;
                         tickCLK <= true;
                         if (j > 7) then
                            j <= 0;
                            sendCMD <= false;
                            state <= EEND;
                         end if;
                      elsif (not sendCMD) and (not sendDAT) then
                         so <= WRIT(j);
                         j <= j + 1;
                         tickCLK <= true;
                         if (j = 7) then
                            j <= 0;
                            sendDAT <= true;
                         end if;
                      elsif (not sendCMD) and (sendDAT) then
                         so <= ADDR(j);
                         j <= j + 1;
                         if (j = 15) then
                             j <= 0;
                             sendCMD <= true;
                         end if;
                      elsif (sendCMD) and (sendDAT) then
                         so <= WDATA(j);
                         j <= j + 1;
                         if (j > 7) then
                            j <= 0;
                            sendDAT <= false;
                            state <= EEND;
                         end if;
                      end if;

                   end if;
                end if;
             elsif (state = RECEIVE) then

                if (rise = '1') then
                   si(j) <= CFG_SI;
                   j <= j + 1;
                   if (j = 7) then
                      j <= 0;
                      state <= EEND;
                   end if;
                end if;

             end if;

          end if;

          end if;

     end process talk;

end architecture ark;
--------------------------------------------------------------------------------





