--------------------------------------------------------------------------------
-- SubModule SettingDST
-- Stores the DST setting
-- The value is initialized upon a reset (RST high) to INIT_DST.
-- When enabled (ENABLE HIGH) a positive edge on UP and DOWN will set
-- and reset the DST value, respectively.
-- If the DST value changes it is indicated by CHG.  
-- Created   3/2/2009 1:16:06 PM
--------------------------------------------------------------------------------
library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity SettingDST is port
   (
     RST      : in    std_logic;
     CLK      : in    std_logic;
     ENABLE   : in    std_logic;
     UP       : in    std_logic;
     DOWN     : in    std_logic;
     CHG      : out   std_logic;
     SET      : out   std_logic;
     INIT_DST : in    std_logic
   );
end SettingDST;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of SettingDST is

-- Signal Declarations
   signal dst_save       : std_logic;
   signal dst_old        : std_logic;
   signal up_old         : std_logic;
   signal down_old       : std_logic;

begin

   -- check for pos edge in up/down
   up_old <= UP when rising_edge (CLK);
   down_old <= DOWN when rising_edge (CLK);

   -- store DST value
   DSTsave: process (CLK) is
   begin
      if rising_edge (CLK) then
         -- initialize on reset
         if RST = '1' then
            dst_save <= INIT_DST;
         else
            -- can only change when selected
            if ENABLE = '1' then
               if (UP = '1') and (up_old = '0') then
                  -- set DST when ^UP
                  dst_save <= '1';
               elsif (DOWN = '1') and (down_old = '0') then
                  -- reset DST when ^DOWN
                  dst_save <= '0';
               end if;
            end if;
         end if;
      end if;
   end process DSTsave;
   SET <= dst_save;

   -- check for change in stored DST value
   dst_old <= dst_save when rising_edge (CLK);
   CHG <= dst_old xor dst_save;

end Structure;
--------------------------------------------------------------------------------
