--------------------------------------------------------------------------------
-- SubModule SettingLeap
-- Stores the leap second settings
-- LEAP is a unsigned value in units of seconds between 0 and 255 (8 bits).
-- The number of leap seconds are initialized upon a reset (RST high) to
-- INIT_LEAP. Upon reset no leap second will be pending.
-- When enabled (ENABLE HIGH) a positive edge on UP and DOWN will increase
-- and decrease the number of leap seconds, respectively.
-- If the leap seconds value changes, it is indicated by CHG.
-- If a leap second is announced, the pending flag (LEAPPEND) can be set by
-- either adding one (ADD_LEAP goes high) or subtracting one (SUB_LEAP).
-- When the end of the year is reached, a leap second will either be added
-- or subtracted automatically. The following second in the new year the
-- value for the numbers of leap seconds will be adjusted accordingly.
-- To clear a pending leap second which has not been added yet, just subtract
-- or add back to zero.
-- Created   3/2/2009 1:37:39 PM
--------------------------------------------------------------------------------
library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity SettingLeap is port
   (
     CLK             : in    std_logic;
     RST             : in    std_logic;
     PREPPS          : in    std_logic; -- one cycle before 1PPS
     ENABLE          : in    std_logic;
     UP              : in    std_logic;
     DOWN            : in    std_logic;
     LEAP            : out   std_logic_vector(7 downto 0);
     CHG             : out   std_logic;
     ADD_LEAP        : in    std_logic;
     SUB_LEAP        : in    std_logic;
     IS_LEAP         : in    std_logic;
     IS_LAST         : in    std_logic;
     LEAPPEND        : out   std_logic;
     LEAPDEL         : out   std_logic;
     INIT_LEAP       : in    std_logic_vector(7 downto 0)

   );
end SettingLeap;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of SettingLeap is

-- Component Declarations
   constant leap_high      : integer := 255;
   constant leap_low       : integer :=   0;

-- Signal Declarations
   signal leap_save        : unsigned (7 downto 0);
   signal leap_set         : unsigned (7 downto 0);
   signal leap_old         : unsigned (7 downto 0);
   signal leap_uneq        : unsigned (7 downto 0);
   signal isleap_old       : std_logic;
   signal leap_pend        : std_logic;
   signal leap_del         : std_logic;
   signal up_old           : std_logic;
   signal down_old         : std_logic;
   signal add_old          : std_logic;
   signal sub_old          : std_logic;
   signal chgchg           : std_logic;
   signal flag             : boolean;

begin

   -- check for pos edge in up/down and add/sub
   up_old <= UP when rising_edge (CLK);
   down_old <= DOWN when rising_edge (CLK);
   add_old <= ADD_LEAP when rising_edge (CLK);
   sub_old <= SUB_LEAP when rising_edge (CLK);

   -- check for last second
   isleap_old <= IS_LEAP when rising_edge (CLK);

   -- store leap second value and pending flags
   LEAPsave: process (CLK) is
   begin
      if rising_edge (CLK) then
         -- initialize upon reset
         if RST = '1' then
--          commented this out for now; ignore INIT_LEAP and initialize to 0
            leap_save <= unsigned (INIT_LEAP);
--            leap_save <= to_unsigned (15,8);
            leap_pend <= '0';
            leap_del <= '0';
         else
            -- last second  of the year: process leap if pending
            if (IS_LEAP = '1') and (isleap_old = '0') then
               if leap_pend = '1' then
                  if leap_del = '0' then
                     -- add a leap second
                     if leap_save /= To_unsigned (leap_high, 8) then
                        leap_save <= leap_save + 1;
                     end if;
                  else
                     -- subtract a leap second
                     if leap_save /= To_unsigned (leap_low, 8) then
                        leap_save <= leap_save - 1;
                     end if;
                  end if;
                  leap_pend <= '0'; -- no longer pending
                  leap_del <= '0';
               end if;
            -- Add a pending leap second
            elsif (ADD_LEAP = '1') and (add_old = '0') then
               if (leap_pend = '1') and (leap_del = '1') then
                  leap_pend <= '0';
                  leap_del <= '0';
               else
                  leap_pend <= '1';
                  leap_del <= '0';
               end if;
            -- Subtract a pending leap second
            elsif (SUB_LEAP = '1') and (sub_old = '0') then
               if (leap_pend = '1') and (leap_del = '0') then
                  leap_pend <= '0';
                  leap_del <= '0';
               else
                  leap_pend <= '1';
                  leap_del <= '1';
               end if;
            -- can only change leap seconds when selected
            elsif ENABLE = '1' then
               -- increase leap seconds when ^UP
               if (UP = '1') and (up_old = '0') then
                  if leap_save /= To_unsigned (leap_high, 8) then
                     leap_save <= leap_save + 1;
                  end if;
               -- decrease leap seconds when ^DOWN
               elsif (DOWN = '1') and (down_old = '0') then
                  if leap_save /= To_unsigned (leap_low, 8) then
                     leap_save <= leap_save - 1;
                  end if;
               end if;
            end if;
         end if;
      end if;
   end process LEAPsave;

   change: process (CLK) is
   begin
      if rising_edge (CLK) then
         if (leap_uneq /= B"00000000") and (not flag) then
            chgchg <= '1';
            flag <= true;
         elsif (chgchg = '1') then
            chgchg <= '0';
         end if;

         if PREPPS = '1' then
            flag <= false;
            leap_set <= leap_save;
            leap_old <= leap_set;
            leap_uneq <= leap_old xor leap_save;
            LEAPPEND <= leap_pend;
            LEAPDEL <= leap_pend and leap_del;
         end if;
      end if;
   end process change;

   LEAP <= std_logic_vector (leap_set);

   CHG <= chgchg;

end Structure;
--------------------------------------------------------------------------------
