--------------------------------------------------------------------------------
-- SubModule SettingTimezone
-- Stores the timezone settings
-- Timezone is a signed value in units of half hours between -24 (-12h) and
-- +28 (+14h). It uses 6 bits (2-complement).
-- The timezone is initialized upon a reset (RST high) to INIT_TZ.
-- When enabled (ENABLE HIGH) a positive edge on UP and DOWN will increase
-- and decrease the timezone value, respectively.
-- If the timezone value changes it is indicated by CHG.
-- Created   3/2/2009 1:22:47 PM
--------------------------------------------------------------------------------
library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity SettingTimezone is port
   (
     RST           : in    std_logic;
     CLK           : in    std_logic;
     ENABLE        : in    std_logic;
     UP            : in    std_logic;
     DOWN          : in    std_logic;
     TZ            : out   std_logic_vector(5 downto 0);
     CHG           : out   std_logic;
     INIT_TZ       : in    std_logic_vector(5 downto 0);

     OFFSET        : out   std_logic_vector(5 downto 0);
     SIGN          : out   std_logic

   );
end SettingTimezone;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of SettingTimezone is

-- Component Declarations
   constant tz_high      : integer := +28; -- 2 * 14
   constant tz_low       : integer := -24; -- 2 * (-12)

-- Signal Declarations
   signal tz_save        : signed (5 downto 0);
   signal tz_old         : signed (5 downto 0);
   signal tz_uneq        : signed (5 downto 0);
   signal up_old         : std_logic;
   signal down_old       : std_logic;

   signal off            : signed (5 downto 0);

begin

   -- check for pos edge in up/down
   up_old <= UP when rising_edge (CLK);
   down_old <= DOWN when rising_edge (CLK);

   -- store timezone value
   TZsave: process (CLK) is
   begin
      if rising_edge (CLK) then
         -- initialize on reset
         if RST = '1' then
            tz_save <= signed (INIT_TZ);
         else
            -- can only change when selected
            if ENABLE = '1' then
               if (UP = '1') and (up_old = '0') then
                  -- increase timezone when ^UP
                  if tz_save /= To_signed (tz_high, 6) then
                     tz_save <= tz_save + 1;
                  end if;
               elsif (DOWN = '1') and (down_old = '0') then
                  -- decrease timezone when ^DOWN
                  if tz_save /= To_signed (tz_low, 6) then
                     tz_save <= tz_save - 1;
                  end if;
               end if;
            end if;
         end if;
      end if;
   end process TZsave;
   TZ <= std_logic_vector (tz_save);

   -- check for change in stored timezone value
   tz_old <= tz_save when rising_edge (CLK);
   tz_uneq <= tz_old xor tz_save;
   CHG <= tz_uneq(0) or tz_uneq(1) or tz_uneq(2) or tz_uneq(3) or
          tz_uneq(4) or tz_uneq(5);

   -- Just some post-processing to make the DIGITAL I/O more readable...

   -- TZ is encoded in 2's complement...
   -- so, subtract the register from zero to get the desired value.
   findOff : process (CLK) is
   begin
        if rising_edge (CLK) then
           if (tz_save(5) = '1') then
              off <= B"000000" - tz_save;
           else
              off <= tz_save;
           end if;
        end if;
   end process findOff;
   OFFSET <= std_logic_vector(off);

   -- Output the sign of the offest in a seperate register
   SIGN <= std_logic(tz_save(5));

end Structure;
--------------------------------------------------------------------------------
