--------------------------------------------------------------------------------
-- If the user wishes to generate a fake IRIG-B test signal which is 100 clock
-- cycles ahead of the reference time, then inherent hardware delay will
-- slow the signal down by a small number of clock cycles, usually around 8. The
-- effective difference will then be 92 clock cycles. To have a signal which is
-- 100 cycles ahead of reference time, it's necessary to generate a signal which
-- actually leads by 108 cycles. A Digital I/O allows the user to input a
-- baseline time delay which compensates for this delay inherent in hardware.

-- This program adds the baseline delay to the desired subsecond difference. For
-- instance, if the user desires a subsecond difference of 100 clock cycles and
-- there is a hardware delay of 8 clock cycles, then the implemented delay must
-- be 108 clock cycles. Alternatively, a negative subsecond difference of 100
-- clock cycles would be implemented as a difference of 92 clock cycles with the
-- other 8 clock cycles made up by inherent hardware delay.
--
-- Ilya Belopolski, January 2012
--------------------------------------------------------------------------------

library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity SubtractDelay is port
   (
       CLK                 : in  std_logic;
       INPUT_DIFF          : in  std_logic_vector (25 downto 0);
       SIGN                : in  std_logic;
       BASELINE_DELAY      : in  std_logic_vector (25 downto 0);
       DIFF                : out std_logic_vector (31 downto 0)
    );
end SubtractDelay;

architecture ark of SubtractDelay is

     signal rawdiff       : signed(31 downto 0);
     signal off           : signed(31 downto 0);

begin

     rawdiff <= signed(B"000000" & INPUT_DIFF);

     finddiff : process (CLK) is
     begin
          if rising_edge (CLK) then
               if (SIGN = '1') then
                  off <= X"00000000" - rawdiff;
               else
                  off <= rawdiff;
               end if;
          end if;
     end process finddiff;

     DIFF <= std_logic_vector(off + signed(B"000000" & BASELINE_DELAY));

end ark;
