--------------------------------------------------------------------------------
-- SubModule TimingLink
--
-- This module generates the status string of a uplink transceiver.
--
-- The interface to the fiber receiver is through a serial interface. This
-- interface has to be initialized every second by applying IN_START, before
-- the first bit is received. Applying IN_CE will read a single bit. When
-- the maximum number of bits have been received IN_DONE goes high. If there
-- was a CRC error in the received data, it will be indicated with the
-- IN_CRCERR bit 1 cycle after DONE goes high.
--
-- The interface to the fiber transmitter is initilaizd with the given
-- GPS time and address, when OUT_START is applied. The first data bit
-- is immediatly available. Whenever OUT_CE is applied, the next bit is
-- shited out. OUT_DONE goes high, when no more data is available.
--
-- Parameters:
--   CLK: clock
--   RST: reset
--   GPS: current GPS time (set from bit input)
--   ADDR: module address (set from bit input)
--   STAT: status string send to the uplink transmitter
--
--   IN_D: serial data from fiber link
--   IN_CE: clock enable for serial input data
--   IN_START: this restarts the stream to the first bit
--             (must come before the first CE)
--   IN_DONE: enough bits have been received
--   IN_CRCERR: indicates a CRC error on the fiber receiver
--              (valid when done goes high)
--
--   OUT_D: serial data to fiber link
--   OUT_CE: clock enable to move to the next bit
--   OUT_START: this restarts the stream to the first bit
--              (must come before the first CE)
--   OUT_DONE: no more data is available
--
-- Created   7/23/2008 DS
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity TimingLink is port
   (
     CLK           : in    std_logic;
     RST           : in    std_logic;

     GPS           : out   std_logic_vector(31 downto 0);
     ADDR          : out   std_logic_vector(31 downto 0);
     A             : out   std_logic_vector(31 downto 0);
     B             : out   std_logic_vector(31 downto 0);
     STAT          : in    std_logic_vector(1023 downto 0);

     IN_D          : in    std_logic;
     IN_CE         : in    std_logic;
     IN_START      : in    std_logic;
     IN_DONE       : out   std_logic;
     IN_CRCERR     : out   std_logic;
     IN_CRCOK      : out   std_logic;

     OUT_D         : out   std_logic;
     OUT_CE        : in    std_logic;
     OUT_START     : in    std_logic;
     OUT_DONE      : out   std_logic
   );
end TimingLink;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TimingLink is

-- Constant Declarations
   constant INPUT_BITS : integer := 128;  -- not including CRC
   constant OUTPUT_BITS : integer := 1024-32; -- not including the CRC
   constant DONE_DURATION : unsigned (31 downto 0) := X"00800000"; -- duration of DONE in CLK units

-- Constant Declarations
   constant CRC32_POLYNOMIAL : std_logic_vector (31 downto 0) := X"04C11DB7";
   constant CRC32_INIT_VALUE : std_logic_vector (31 downto 0) := X"FFFFFFFF";
   constant CRC32_MATCH_VALUE: std_logic_vector (31 downto 0) := X"C704DD7B";

-- Component Declarations
   component ucrc_ser is
      generic (
         POLYNOMIAL : std_logic_vector;
         INIT_VALUE : std_logic_vector;
         SYNC_RESET : integer range 0 to 1 := 0);  -- use synchronous reset
      port (
         clk_i   : in  std_logic;          -- clock
         rst_i   : in  std_logic;          -- init CRC
         clken_i : in  std_logic;          -- clock enable
         data_i  : in  std_logic;          -- data input
         flush_i : in  std_logic;          -- flush crc
         match_o : out std_logic;          -- CRC match flag
         crc_o   : out std_logic_vector(POLYNOMIAL'length - 1 downto 0));  -- CRC output
   end component;


-- Type Declarations
   type in_state_type    is (SHIFT_D, SHIFT_WAIT, SHIFT_DONE);
   type out_state_type   is (SHIFT_D, SHIFT_CRC, SHIFT_DONE);

-- Signals
   signal in_din         : std_logic_vector (INPUT_BITS+32-1 downto 0);
   signal in_crcmatch    : std_logic;
   signal in_count       : unsigned (15 downto 0);
   signal in_state       : in_state_type;
--   signal in_done_count  : unsigned (31 downto 0);
   signal in_end          : std_logic;

   signal out_reg        : std_logic_vector (OUTPUT_BITS-1 downto 0);
   signal out_data       : std_logic;
   signal out_crc_flush  : std_logic;
   signal out_crc        : std_logic_vector (31 downto 0);
   signal out_count      : unsigned (15 downto 0);
   signal out_state      : out_state_type;

begin

   -----------------------------------------------------------------------------
   -- SHIFT FROM INPUT FIBER
   -----------------------------------------------------------------------------

   -- input CRC generator
   input_crc: ucrc_ser
      generic map (
         POLYNOMIAL => CRC32_POLYNOMIAL,
         INIT_VALUE => CRC32_INIT_VALUE,
         SYNC_RESET => 0)
      port map (
         clk_i   => CLK,
         rst_i   => IN_START,
         clken_i => IN_CE,
         data_i  => IN_D,
         flush_i => '0',
         match_o => in_crcmatch,
         crc_o   => open
      );

   -- input state machine
   in_fsm: process (CLK) is
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            in_end <= '0';
            in_state <= SHIFT_DONE;
            in_count <= To_Unsigned (0, 16);
         elsif IN_START = '1' then
            in_state <= SHIFT_D;
            in_count <= To_Unsigned (INPUT_BITS+32-1, 16);
         else
            case in_state is
               when SHIFT_D =>
                  if IN_CE = '1' then
                     in_din <= in_din (INPUT_BITS+32-2 downto 0) & IN_D;
                     in_count <= in_count - 1;
                     if in_count = To_Unsigned (0, 16) then
                        GPS   <= in_din (INPUT_BITS+32-2 downto INPUT_BITS+32-33);
                        ADDR  <= in_din (INPUT_BITS+32-34 downto INPUT_BITS+32-65);
                        A     <= in_din (INPUT_BITS+32-66 downto INPUT_BITS+32-97);
                        B     <= in_din (INPUT_BITS+32-98 downto INPUT_BITS+32-129);
                        in_count <= To_Unsigned (15, 16);
                        in_state <= SHIFT_WAIT;
                     end if;
                  end if;
               when SHIFT_WAIT =>
                  in_count <= in_count - 1;
                  if in_count = To_Unsigned (0, 16) then
                     in_end <= '1';
                     in_state <= SHIFT_DONE;
                  end if;
               when SHIFT_DONE =>
                  in_end <= '0';
                  -- nothing
            end case;
         end if;
      end if;
   end process in_fsm;
   IN_DONE <= in_end;
   IN_CRCERR <= in_end and not in_crcmatch;
   IN_CRCOK  <= in_end and in_crcmatch;

   -----------------------------------------------------------------------------
   -- SHIFT TO OUTPUT FIBER
   -----------------------------------------------------------------------------

   -- output CRC generator
   output_crc: ucrc_ser
      generic map (
         POLYNOMIAL => CRC32_POLYNOMIAL,
         INIT_VALUE => CRC32_INIT_VALUE,
         SYNC_RESET => 0)
      port map (
         clk_i   => CLK,
         rst_i   => OUT_START,
         clken_i => OUT_CE,
         data_i  => out_data,
         flush_i => out_crc_flush,
         match_o => open,
         crc_o   => out_crc
      );

   -- output bits
   out_data <= out_reg(out_reg'high);
   OUT_D <= out_data when (out_state /= SHIFT_CRC) else out_crc(31);

   -- output state machine
   out_fsm: process (CLK) is
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            out_reg <=(others => '0');
            out_state <= SHIFT_DONE;
         elsif OUT_START = '1' then
            out_reg <= STAT (STAT'high downto STAT'high-OUTPUT_BITS+1);
            out_state <= SHIFT_D;
            out_crc_flush <= '0';
            out_count <= To_Unsigned (OUTPUT_BITS - 1, 16);
         else
            case out_state is
               when SHIFT_D =>
                  if OUT_CE = '1' then
                     out_reg <= out_reg (OUTPUT_BITS-2 downto 0) & '0';
                     out_count <= out_count - 1;
                     if out_count = To_Unsigned (0, 16) then
                        out_state <= SHIFT_CRC;
                        out_crc_flush <= '1';
                        out_count <= To_Unsigned (31, 16);
                     end if;
                  end if;
               when SHIFT_CRC =>
                  if OUT_CE = '1' then
                     out_count <= out_count - 1;
                     if out_count = To_Unsigned (0, 16) then
                        out_state <= SHIFT_DONE;
                     end if;
                  end if;
               when SHIFT_DONE =>
                  out_crc_flush <= '0';
            end case;
         end if;
      end if;
   end process out_fsm;

   OUT_DONE <= '1' when (out_state = SHIFT_DONE)
                  else '0';

end architecture Structure;

