--------------------------------------------------------------------------------
-- SubModule UART_GPS
-- Created   12/19/2007 4:58:13 PM
-- Modified 03/08/2008
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity UART is port
   (
     CLK        : in    std_logic;
     RST        : in    std_logic;
     DIV        : in    std_logic_vector(15 downto 0);

     RX         : in    std_logic;
     TX         : out   std_logic;

     LD         : in    std_logic;
     DIN        : in    std_logic_vector(7 downto 0);
     BUSY       : out   std_logic;

     RDY        : out   std_logic;
     ERR        : out   std_logic;
     DOUT       : out   std_logic_vector(7 downto 0)
   );
end UART;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of UART is

-- Component Declarations

-- Constant Declarations
   constant parity    : boolean := false;
   constant NDBits    : integer := 8;
   constant even      : std_logic := '0';
   constant odd       : std_logic := '1';

-- Type Declarations
   type TxFSM_state is (idle, Load_Tx, Shift_Tx, Stop_Tx);
   type RxFSM_state is (idle, Start_Rx, Edge_Rx, Shift_Rx, Stop_Rx, RxOVF);

-- Signal Declarations
   signal Top16       : std_logic;
   signal Top16Old    : std_logic;
   signal TopTx       : std_logic;
   signal TopRx       : std_logic;
   signal ClrDiv      : std_Logic;
   signal Div16       : unsigned (15 downto 0);
   signal ClkDiv      : unsigned (3 downto 0);
   signal RxDiv       : unsigned (3 downto 0);

   signal Tx_Reg      : std_logic_vector (9 downto 0);
   signal TxBitCnt    : unsigned (3 downto 0);
   signal TxFSM       : TxFSM_state;
   signal TxBusy      : std_logic;
   signal RegDin      : std_logic_vector (7 downto 0);

   signal Rx_Reg      : std_logic_vector (7 downto 0);
   signal RxBitCnt    : unsigned (3 downto 0);
   signal RxFSM       : RxFSM_state;
   signal RxRdyi      : std_logic;
   signal RxErr       : std_logic;

   signal RxOld       : std_logic; -- New signal, for distinguishing start-bits
                                   -- from "no-input" state

-- functions
   function make_parity (d : in std_logic_vector (7 downto 0);
                         p : in std_logic) return std_logic is
   begin
      return d(0) xor d(1) xor d(2) xor d(3) xor d(4) xor
             d(5) xor d(6) xor d(7) xor p;
   end function make_parity;

begin
   -- --------------------------
   -- Clk16 Clock Generation
   -- --------------------------
   process (RST, CLK)
   begin
      if RST='1' then
         Top16 <= '0';
         Div16 <= To_Unsigned (0, 16);
      elsif rising_edge(CLK) then
         Top16 <= '0';
         if Div16 = unsigned (DIV) then
            Div16 <= To_Unsigned (0, 16);
            Top16 <= '1';
         else
            Div16 <= Div16 + 1;
         end if;
      end if;
   end process;

   -- --------------------------
   -- Tx Clock Generation
   -- --------------------------
   process (RST, CLK)
   begin
      if RST='1' then
         TopTx <= '0';
         ClkDiv <= (others=>'0');
      elsif rising_edge(CLK) then
         TopTx <= '0';
         if Top16='1' then
            ClkDiv <= ClkDiv + 1;
            if ClkDiv = To_Unsigned (15, 4) then
               TopTx <= '1';
            end if;
         end if;
      end if;
   end process;

   -- ------------------------------
   -- Rx Sampling Clock Generation
   -- ------------------------------
   process (RST, CLK)
   begin
      if RST='1' then
         TopRx <= '0';
         RxDiv <= To_Unsigned (0, 4);
      elsif rising_edge(CLK) then
         TopRx <= '0';
         if ClrDiv='1' then
            RxDiv <= To_Unsigned (0, 4);
         elsif Top16='1' then
            if RxDiv = To_Unsigned (7, 4) then
               RxDiv <= To_Unsigned (0, 4);
               TopRx <= '1';
            else
               RxDiv <= RxDiv + 1;
            end if;
         end if;
      end if;
   end process;

   -- --------------------------
   -- Transmit State Machine
   -- --------------------------
   Tx_FSM: process (RST, CLK)
   begin
      if RST='1' then
         Tx_Reg <= (others => '1');
         TxBitCnt <= To_Unsigned (0, 4);
         TxFSM <= idle;
         TxBusy <= '0';
         RegDin <= (others=>'0');
      elsif rising_edge(CLK) then
         TxBusy <= '1'; -- except when explicitly '0'
         case TxFSM is
            when Idle =>
               if LD='1' then
                  -- latch the input data immediately.
                  RegDin <= Din;
                  TxBusy <= '1';
                  TxFSM <= Load_Tx;
               else
                  TxBusy <= '0';
               end if;
            when Load_Tx =>
               if TopTx='1' then
                  TxFSM <= Shift_Tx;
                  if parity then
                     -- start + data + parity
                     TxBitCnt <= To_Unsigned (NDBits + 2, 4);
                     Tx_Reg <= make_parity (RegDin, even) & Din & '0';
                  else
                     TxBitCnt <= To_Unsigned (NDBits + 1, 4); -- start + data
                     Tx_reg <= '1' & RegDin & '0';
                  end if;
               end if;
            when Shift_Tx =>
               if TopTx='1' then
                  TxBitCnt <= TxBitCnt - 1;
                  Tx_reg <= '1' & Tx_reg (Tx_reg'high downto 1);
                  if TxBitCnt = To_Unsigned (1, 4) then
                     TxFSM <= Stop_Tx;
                  end if;
               end if;
            when Stop_Tx =>
               if TopTx='1' then
                  TxFSM <= Idle;
               end if;
            when others =>
               TxFSM <= Idle;
         end case;
      end if;
   end process;
   TX <= Tx_Reg(0);
   BUSY <= TxBusy;

   -- ------------------------
   -- RECEIVE State Machine
   -- ------------------------
   Top16Old <= Top16 when rising_edge (CLK);
   Rx_FSM: process (RST, CLK)
   begin
      if RST='1' then
         Rx_Reg <= (others => '0');
         Dout <= (others => '0');
         RxBitCnt <= To_Unsigned (0, 4);
         RxFSM <= Idle;
         RxRdyi <= '0';
         ClrDiv <= '0';
         RxErr <= '0';
      elsif rising_edge(CLK) then
         ClrDiv <= '0'; -- default value
         -- reset error when a word has been received Ok:
         if RxRdyi='1' then
            RxErr <= '0';
            RxRdyi <= '0';
         end if;
         case RxFSM is
            when Idle => -- wait on start bit
               RxBitCnt <= To_Unsigned (0, 4);
               RxOld <= RX;
               if Top16Old='1' then
                  if RX='0' then
                     RxFSM <= Start_Rx;
                     ClrDiv <='1'; -- Synchronize the divisor
                  end if; -- else false start, stay in Idle
               end if;

            when Start_Rx => -- wait on first data bit
               if TopRx = '1' then
                  if RX = '1' then -- framing error
                     RxFSM <= RxOVF;
--                     report "Start bit error." severity note;
                  else
                     RxFSM <= Edge_Rx;
                  end if;
               end if;

            when Edge_Rx => -- should be near Rx edge
               if TopRx = '1' then
                  if RxBitCnt = NDbits then
                     RxFSM <= Stop_Rx;
                  else
                     RxFSM <= Shift_Rx;
                  end if;
               end if;

            when Shift_Rx => -- Sample data !
               if TopRx = '1' then
                  RxBitCnt <= RxBitCnt + 1;
                  -- shift right :
                  Rx_Reg <= RX & Rx_Reg (Rx_Reg'high downto 1);
                  RxFSM <= Edge_Rx;
               end if;

            when Stop_Rx => -- during Stop bit
               if TopRx = '1' then
                  Dout <= Rx_reg;
                  if RxOld = '0' and RX = '1' then
                     RxRdyi <='1';   -- set RDY only when start-bit was present
                  end if;
                  RxFSM <= Idle;
--                  assert (true)
--                     report "Character received in decimal is : "
--                        & integer'image(to_integer(unsigned(Rx_Reg)))
--                        severity note;
               end if;

            when RxOVF => -- Overflow / Error
               RxErr <= '1';
               if Rx='1' then
                  RxFSM <= Idle;
               end if;
         end case;
      end if;
   end process;
   ERR <= RxErr;
   RDY <= RxRdyi;

end Structure;
--------------------------------------------------------------------------------
