--------------------------------------------------------------------------------
-- EEPROM Writer
-- Reads & writes LEAP, TZ, & DST information from & to the EEPROM
--
-- Ilya Belopolski
-- 23 August, 2010
--------------------------------------------------------------------------------

library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity WriteEEPROM is port
    (
      CLK          : in std_logic;

      -- EEPROM Driver Interface --
      RST          : out std_logic; -- The driver will read/write on a rising edge of RST, if RDY = '1';
      RDY          : in std_logic;
      RDWR         : out std_logic; -- Toggles between read ('0') and write ('1');
      ADDR         : out std_logic_vector (15 downto 0);
      WDATA        : out std_logic_vector (7 downto 0);
      RDATA        : in std_logic_vector (7 downto 0);
      SYSRST       : in std_logic;

      -- LEAP, TZ, DST Interface --
      SETRST       : out std_logic;
      LEAP         : in std_logic_vector (7 downto 0);
      TZ           : in std_logic_vector (5 downto 0);
      DST          : in std_logic;
      INIT_LEAP    : out std_logic_vector (7 downto 0);
      INIT_TZ      : out std_logic_vector (5 downto 0);
      INIT_DST     : out std_logic;
      LEAP_CHG     : in std_logic;
      TZ_CHG       : in std_logic;
      DST_CHG      : in std_logic

    );
end WriteEEPROM;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture ark of WriteEEPROM is

     signal initialize   : boolean := false;
     signal pause        : boolean := false;
     signal count        : integer := 0;

     signal started      : boolean := false;
     signal finished     : boolean := false;
     signal done         : integer := 0;

     signal RW           : std_logic;
     signal AD           : std_logic_vector (15 downto 0);
     signal CRST         : std_logic; -- Chip reset buffer
     signal SRST         : std_logic; -- LEAP, TZ, DST reset buffer
     signal WD           : std_logic_vector (7 downto 0);

     -- These hold the data read in from the EEPROM
     signal InfoLEAP     : std_logic_vector (7 downto 0);
     signal InfoTZ       : std_logic_vector (7 downto 0);
     signal InfoDST      : std_logic_vector (7 downto 0);

     -- Keep track of which registers need to be updated
     signal updateLEAP   : boolean := false;
     signal updateTZ     : boolean := false;
     signal updateDST    : boolean := false;

     -- Keep track of which signals are being processed
     signal processLEAP  : boolean := false;
     signal processTZ    : boolean := false;
     signal processDST   : boolean := false;

begin

     RDWR <= RW;
     ADDR <= AD;
     RST <= CRST;
     WDATA <= WD;

     SETRST <= SRST;
     INIT_LEAP <= InfoLEAP;
     INIT_TZ <= InfoTZ (5 downto 0);
     INIT_DST <= InfoDST (0);

     Junk : process (CLK) is
     begin
          if rising_edge (CLK) then

          if (SYSRST = '1') then

             initialize <= false;
             pause <= false;
             count <= 0;

             started <= false;
             finished <= false;
             done <= 0;

             updateLEAP <= false;
             updateTZ <= false;
             updateDST <= false;

             processLEAP <= false;
             processTZ <= false;
             processDST <= false;

          else

            -- This block loads in LEAP, TZ, & DST data from the EEPROM --
            if (not initialize) and (not pause) then

                RW <= '0';
                SRST <= '1';

                if (done = 0) or (done = 3) then
                   AD <= X"0001";
                elsif (done = 1) then
                   AD <= X"0010";
                elsif (done = 2) then
                   AD <= X"0100";
                end if;

                if (RDY = '1') and (not started) then
                   CRST <= '1';
                   started <= true;
                elsif (RDY = '0') and (started) then
                   finished <= true;
                elsif (RDY = '1') and (finished) then
                   CRST <= '0';
                   started <= false;
                   finished <= false;
                   if (done = 0) then
                      InfoLEAP <= RDATA;
                   elsif (done = 1) then
                      InfoTZ <= RDATA;
                   elsif (done = 2) then
                      InfoDST <= RDATA;
                   elsif (done = 3) then
                      InfoLEAP <= RDATA;
                      pause <= true;
                   end if;
                   done <= done + 1;
                end if;

             -- This block writes any changes in LEAP, TZ, & DST to the EEPROM --
             elsif initialize then

                RW <= '1';
                SRST <= '0';

                if (LEAP_CHG = '1') then
                   updateLEAP <= true;
                end if;
                if (TZ_CHG = '1') then
                   updateTZ <= true;
                end if;
                if (DST_CHG = '1') then
                   updateDST <= true;
                end if;

                if updateLEAP then
                   AD <= X"0001";
                   WD <= LEAP;
                elsif updateTZ then
                   AD <= X"0010";
                   WD <= B"00" & TZ;
                elsif updateDST then
                   AD <= X"0100";
                   WD <= B"0000000" & DST;
                end if;

                if (RDY = '1') and (not started) then
                   if (updateLEAP) then
                       processLEAP <= true;
                       CRST <= '1';
                       started <= true;
                   elsif (updateTZ) then
                       processTZ <= true;
                       CRST <= '1';
                       started <= true;
                   elsif (updateDST) then
                       processDST <= true;
                       CRST <= '1';
                       started <= true;
                   end if;
                elsif (RDY = '0') and (started) then
                   finished <= true;
                elsif (RDY = '1') and (finished) then
                   CRST <= '0';
                   started <= false;
                   finished <= false;
                   if (processLEAP) then
                      updateLEAP <= false;
                      processLEAP <= false;
                   elsif (processTZ) then
                      updateTZ <= false;
                      processTZ <= false;
                   elsif (processDST) then
                      updateDST <= false;
                      processDST <= false;
                   end if;
                end if;

             elsif pause then
                count <= count + 1;

                if (count = 10) then
                   pause <= false;
                   initialize <= true;
                end if;

             end if;

          end if;

          end if;
     end process Junk;

end architecture ark;


