--------------------------------------------------------------------------------
-- SubModule BCDConv
-- Created   12/19/2007 4:58:13 PM
-- Modified 03/08/2008
-- Binary to BCD conversion
-- M represents the number of binary digits in the input.
-- N represents the number of BCD digits in the final result.
-- Everytime a bit is shifted in the old value is doubled and the new bit added.
-- Whenever the resulting sum becomes larger than 10, a carry on is sent to
-- the next more significant digit. By shifting the binary value into the
-- carry in of the least significant digit, we reconstruct the original value
-- in BCD format. The conversion takes as many cycles as there are bits in the
-- binary number.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.Numeric_Std.all;

entity BCDConv is
   generic (
      M           : positive;      -- number of binary digits
      N           : positive       -- number of BCD digits
   );
   port (
      CLK         : in  std_logic;
      RST         : in  std_logic;
      INIT        : in  std_logic;  -- initialize conversion
      BUSY        : out std_logic;  -- indicates busy with conversion
      DIN         : in  std_logic_vector (M - 1 downto 0);    -- binary input
      DOUT        : out std_logic_vector (4 * N - 1 downto 0); -- BCD result
      OVF         : out std_logic   -- overflow
);
end;

architecture Structure of BCDConv is

-- Component Declarations
   component Digit
      port (
        CLK         : in std_logic;      -- working clock
        RST         : in std_logic;      -- clear registers asynchronously
        INIT        : in std_logic;      -- initialise the BCD conversion
        ENABLE      : in std_logic;      -- enable doubling
        CARRYIN     : in std_logic;      -- modulus in from less significant digit
        CARRYOUT    : out std_logic;     -- modulus out to more significant digit
        D           : out std_logic_vector(3 downto 0) -- BCD output
      );
   end component;

-- Signal Declarations
   signal carry       : std_logic_vector (N downto 0);
   signal shift       : std_logic_vector (M - 1 downto 0);
   signal count       : unsigned (7 downto 0);
   signal enable      : std_logic;
   signal delay       : std_logic;
   signal overflow    : std_logic;
   signal bcd         : std_logic_vector (4 * N - 1 downto 0);

begin

   -- Generate N linked digits
   gen_digit : for i in 1 to N generate
      d1: Digit
        port map (
           CLK => CLK,
           RST => RST,
           INIT => INIT,
           ENABLE => enable,
           CARRYIN => carry (i - 1),
           CARRYOUT => carry (i),
           D => bcd (i * 4 - 1 downto i * 4 - 4)
        );
   end generate gen_digit;
   DOUT <= bcd;

  -- initialize carry chain
  carry (0) <= shift (M - 1);

  -- generate clock enable
  conv: process (CLK, RST) is
  begin
     if RST = '1' then
        count <= (others => '0');
        enable <= '0';
        delay <= '0';
     elsif rising_edge (CLK) then
        if INIT = '1' then
           count <= To_Unsigned (M-1, 8);
           enable <= '0';
           delay <= '1';
           shift <= DIN;
        elsif delay = '1' then
           delay <= '0';
           enable <= '1';
        elsif count /= To_Unsigned (0, 8) then
           count <= count - 1;
           shift <= shift (M - 2 downto 0) & '0';
           delay <= '1';
           enable <= '0';
        else
           enable <= '0';
        end if;
     end if;
  end process conv;
  BUSY <= INIT or enable or delay;

  -- overflow
  overflw: process (CLK, RST) is
  begin
     if RST = '1' then
        overflow <= '0';
     elsif rising_edge (CLK) then
        if INIT = '1' then
           overflow <= '0';
        else
           overflow <= overflow or carry (N);
        end if;
     end if;
  end process overflw;
  OVF <= overflow;

end architecture Structure;

