--------------------------------------------------------------------------------
-- SubModule Digit
-- Created   12/19/2007 4:58:13 PM
-- Modified 03/08/2008
-- One digit in the binary to BCD conversion
-- Everytime a bit is shifted in the old value is doubled and the new bit added.
-- Whenever the resulting sum becomes larger than 10, a carry on is sent to
-- the next more significant digit. By shifting the binary value into the
-- carry in of the least significant digit, we reconstruct the original value
-- in BCD format. The conversion takes as many cycles as there are bits in the
-- binary number.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;

entity Digit is port
   (
     CLK         : in std_logic;      -- working clock
     RST         : in std_logic;      -- clear registers asynchronously
     INIT        : in std_logic;      -- initialise the BCD conversion
     ENABLE      : in std_logic;      -- enable doubling
     CARRYIN     : in std_logic;      -- modulus in from less significant digit
     CARRYOUT    : out std_logic;     -- modulus out to more significant digit
     D           : out std_logic_vector(3 downto 0) -- BCD output
   );
end;

architecture Structure of Digit is

-- Signal Declarations
   signal D_int       : std_logic_vector(3 downto 0); -- internal digit
   signal next_D      : std_logic_vector(2 downto 0); -- next digit
   signal next_carry  : std_logic;                    -- next carry

begin

   -- Calculate the shift in the BCD register. Numbers between
   -- 0 and 4 inclusive are doubled, by shifting by 1.
   -- Numbers from  5 to 9 inclusive get mapped to 10, 12, 14,
   -- 16, 18. This gives an modout of 1 (i.e. a carry to the
   -- next digit), and the values 0, 2, 4, 6, 8.
   BCDDoubler : process (D_int)
   begin
      case D_int is
         when "0000" =>
                              -- Input 10s   Units    Result
            next_D <= "000";  --  0     0     0         0
         when "0001" =>
            next_D <= "001";  --  1     0     2         2
         when "0010" =>
            next_D <= "010";  --  2     0     4         4
         when "0011" =>
            next_D <= "011";  --  3     0     6         6
         when "0100" =>
            next_D <= "100";  --  4     0     8         8
         when "0101" =>
            next_D <= "000";  --  5     1     0         10
         when "0110" =>
            next_D <= "001";  --  6     1     2         12
         when "0111" =>
            next_D <= "010";  --  7     1     4         14
         when "1000" =>
            next_D <= "011";  --  8     1     6         16
         when "1001" =>
            next_D <= "100";  --  9     1     8         18
         when others =>
            next_D <= (others => '-');
      end case;
   end process BCDDoubler;

   -- if the numbers are greater than 5, we should generate a
   -- carry out (modulus out) to the next digit.
   CarryGen : process (D_int)
   begin
      case D_int is
         when "0101" | "0110" | "0111" | "1000" | "1001" =>
            next_carry <= '1';
         when others =>
            next_carry <= '0';
      end case;
   end process CarryGen;

   -- When Init is high, we force ModOut to 0
   CARRYOUT <= next_carry and (not INIT);

   ShiftReg : process (CLK)
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            D_int <= (others => '0');
         elsif INIT = '1' then
            D_int <= (others => '0');
         elsif ENABLE = '1' then
            D_int <= Next_D & CARRYIN;
         end if;
      end if;
   end process ShiftReg;

  D <= D_int;

end architecture Structure;
